!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: INTERPOLATIONS
!!
!!  This module provides subroutine to interpolate variables and reconstruct
!!  Riemann states.
!!
!!******************************************************************************
!
module interpolations

! module variables are not implicit by default
!
  implicit none

! module parameters
!
  real, save :: rad   = 5.0d-01
  real, save :: eps   = epsilon(rad)
  real, save :: kappa = 1.0d+00
  real, save :: kbeta = 1.0d+00

! by default everything is public
!
  public

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! subroutine INITIALIZE_INTERPOLATIONS:
! ------------------------------------
!
!   Subroutine initializes the interpolation module by reading the module
!   parameters.
!
!===============================================================================
!
  subroutine initialize_interpolations()

! include external procedures
!
    use parameters, only : get_parameter_real

! local variables are not implicit by default
!
    implicit none

! local variables
!
    real :: cfl = 0.5d+00
!
!-------------------------------------------------------------------------------
!
! obtain the interpolation coefficients
!
    call get_parameter_real("rad"  , rad  )
    call get_parameter_real("eps"  , eps  )
    call get_parameter_real("kappa", kappa)
    call get_parameter_real("kbeta", kbeta)
    call get_parameter_real("cfl"  , cfl  )

! calculate κ = (1 - ν) / ν
!
    kappa = min(kappa, (1.0d+00 - cfl) / cfl)

!-------------------------------------------------------------------------------
!
  end subroutine initialize_interpolations
!
!===============================================================================
!
! subroutine RECONSTRUCT:
! ----------------------
!
!   Subroutine calls a reconstruction procedure, depending on the compilation
!   flag SPACE, in order to interpolate the left and right states from their
!   cell integrals.  These states are required by any approximate Riemann
!   solver.
!
!   Arguments:
!
!     n  - the length of the input vector;
!     h  - the spatial step; this is required for some reconstruction methods;
!     f  - the input vector of cell averaged values;
!     fl - the left side state reconstructed for location (i+1/2);
!     fr - the right side state reconstructed for location (i+1/2);
!     s  - the input array of shock indicators;
!
!===============================================================================
!
  subroutine reconstruct(n, h, f, fl, fr, s)

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer              , intent(in)  :: n
    real                 , intent(in)  :: h
    real   , dimension(n), intent(in)  :: f
    real   , dimension(n), intent(out) :: fl, fr
    logical, dimension(n), intent(in)  :: s
!
!-------------------------------------------------------------------------------
!
! call the reconstruction subroutine depending on the SPACE flag
!
#ifdef TVD
    call tvd  (n, h, f(:), fl(:), fr(:), s(:))
#endif /* TVD */
#ifdef CENO3
    call ceno3(n, h, f(:), fl(:), fr(:), s(:))
#endif /* CENO3 */
#ifdef LIMO3
    call limo3(n, h, f(:), fl(:), fr(:), s(:))
#endif /* LIMO3 */
#ifdef WENO3
    call weno3(n, h, f(:), fl(:), fr(:), s(:))
#endif /* WENO3 */
#if defined WENO5 || defined WENOZ
    call wenoz(n, h, f(:), fl(:), fr(:), s(:))
#endif /* WENO5 | WENOZ */
#ifdef MP
    call mp   (n, h, f(:), fl(:), fr(:), s(:))
#endif /* MP */

#ifdef CLIP_EXTREMA
! correct the reconstruction near discontinuities in order to improve
! the stability of scheme
!
    call correct_reconstruction(n, f(:), fl(:), fr(:))
#endif /* CLIP_EXTREMA */

!-------------------------------------------------------------------------------
!
  end subroutine reconstruct
!
#ifdef TVD
!
!===============================================================================
!
! subroutine TVD:
! --------------
!
!   Subroutine reconstructs the interface states using the second order TVD
!   method with a limiter selected through a compilation flag LIMITER.
!
!   Arguments are described in subroutine reconstruct().
!
!===============================================================================
!
  subroutine tvd(n, h, f, fl, fr, s)

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer              , intent(in)  :: n
    real                 , intent(in)  :: h
    real   , dimension(n), intent(in)  :: f
    real   , dimension(n), intent(out) :: fl, fr
    logical, dimension(n), intent(in)  :: s

! local variables
!
    integer            :: i, im1
    real               :: df
    real, dimension(n) :: dl
!
!-------------------------------------------------------------------------------
!
! calculate limited derivative
!
    call limiter(n, f(:), dl(:))

! calculate the left- and right-side interface interpolations
!
    do i = 1, n

! prepare neighbour indices
!
      im1 = max(1, i - 1)

! update the left and right-side interpolation states
!
#ifdef SHOCK_DETECTION
      if (s(i)) then

        fl(i) = f(i)
        fr(i) = f(i)

      else
#endif /* SHOCK_DETECTION */
        df = 0.5d+00 * dl(i)

        fl(i) = f(i) + df
        fr(i) = f(i) - df
#ifdef SHOCK_DETECTION
      end if ! shock vicinity
#endif /* SHOCK_DETECTION */

! shift the right state
!
      fr(im1) = fr(i)

    end do

! update the interpolation of the first and last points
!
    fl(1) = f(1)
    fr(n) = f(n)

!-------------------------------------------------------------------------------
!
  end subroutine tvd
!
#endif /* TVD */
#ifdef CENO3
!
!===============================================================================
!
! subroutine CENO3:
! ----------------
!
!   Subroutine reconstructs the interface states using the third order
!   Convex Essentially Non-Oscillatory (CENO) method.
!
!   Arguments are described in subroutine reconstruct().
!
!   References:
!
!     [1] Londrillo & Del Zanna, 2000, The Astrophysical Journal, 530, 508
!
!===============================================================================
!
  subroutine ceno3(n, h, f, fl, fr, s)

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer              , intent(in)  :: n
    real                 , intent(in)  :: h
    real   , dimension(n), intent(in)  :: f
    real   , dimension(n), intent(out) :: fl, fr
    logical, dimension(n), intent(in)  :: s

! local variables
!
    integer               :: i, im2, im1, ip1, ip2
    integer, dimension(1) :: ii
    real   , dimension(3) :: q, d
    real   , dimension(n) :: dl
    real                  :: df

! interpolation coefficients
!
    real, parameter :: a11 =   1.0d0 / 3.0d0, a12 = - 7.0d0 / 6.0d0            &
                     , a13 =  11.0d0 / 6.0d0
    real, parameter :: a21 = - 1.0d0 / 6.0d0, a22 =   5.0d0 / 6.0d0            &
                     , a23 =   1.0d0 / 3.0d0
    real, parameter :: a31 =   1.0d0 / 3.0d0, a32 =   5.0d0 / 6.0d0            &
                     , a33 = - 1.0d0 / 6.0d0
!
!-------------------------------------------------------------------------------
!
! calculate the limited derivative
!
    call limiter(n, f(:), dl(:))

! iterate along the vector indices
!
    do i = 1, n

! prepare neighbour indices
!
      im1 = max(1, i - 1)

! update the left and right-side interpolation states
!
#ifdef SHOCK_DETECTION
      if (s(i)) then

        fl(i) = f(i)
        fr(i) = f(i)

      else
#endif /* SHOCK_DETECTION */

! calculate values for the left and right states using the 1st order TVD
! interpolation
!
        df = 0.5d0 * dl(i)

        fl(i) = f(i) + df
        fr(i) = f(i) - df

! prepare indices
!
        im2 = max(1, i - 2)
        ip1 = min(n, i + 1)
        ip2 = min(n, i + 2)

! calculate the 2nd order interpolation of the left state on stencils
! (-2,-1,0), (-1,0,1), and (0,1,2)
!
        q(1) = a11 * f(im2) + a12 * f(im1) + a13 * f(i  )
        q(2) = a21 * f(im1) + a22 * f(i  ) + a23 * f(ip1)
        q(3) = a31 * f(i  ) + a32 * f(ip1) + a33 * f(ip2)

! calculate the interpolation difference between both orders for the left
! interface values
!
        d(1) = q(1) - fl(i)
        d(2) = q(2) - fl(i)
        d(3) = q(3) - fl(i)

! select the closest stencil for both interface values
!
        if(minval(d) * maxval(d) > 0.0d0) then
          d(2)  = 0.7d0 * d(2)
          ii    = minloc(abs(d))
          fl(i) = q(ii(1))
        end if

! calculate the 2nd order interpolation of the right state on stencils
! (-2,-1,0), (-1,0,1), and (0,1,2)
!
        q(1) = a11 * f(ip2) + a12 * f(ip1) + a13 * f(i  )
        q(2) = a21 * f(ip1) + a22 * f(i  ) + a23 * f(im1)
        q(3) = a31 * f(i  ) + a32 * f(im1) + a33 * f(im2)

! calculate the interpolation difference between both orders for the right
! interface values
!
        d(1) = q(1) - fr(i)
        d(2) = q(2) - fr(i)
        d(3) = q(3) - fr(i)

! select the closest stencil for both interface values
!
        if(minval(d) * maxval(d) > 0.0d0) then
          d(2)  = 0.7d0 * d(2)
          ii    = minloc(abs(d))
          fr(i) = q(ii(1))
        end if

#ifdef SHOCK_DETECTION
      end if ! shock vicinity
#endif /* SHOCK_DETECTION */

! shift the right state
!
      fr(im1) = fr(i)

    end do

! update the interpolation of the first and last points
!
    fl(1) = f (1)
    fr(n) = fl(n)

!-------------------------------------------------------------------------------
!
  end subroutine ceno3
#endif /* CENO3 */
#ifdef LIMO3
!
!===============================================================================
!
! subroutine LIMO3:
! ----------------
!
!   Subroutine reconstructs the interface states using the third order method
!   with a limiter function LimO3.
!
!   Arguments are described in subroutine reconstruct().
!
!   References:
!
!     [1] Cada, M. & Torrilhon, M.,
!         "Compact third-order limiter functions for finite volume methods",
!         Journal of Computational Physics, 2009, 228, 4118-4145
!     [2] Mignone, A., Tzeferacos, P., & Bodo, G.,
!         "High-order conservative finite divergence GLM-MHD schemes for
!          cell-centered MHD",
!         Journal of Computational Physics, 2010, 229, 5896-5920
!
!===============================================================================
!
  subroutine limo3(n, h, f, fl, fr, s)

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer              , intent(in)  :: n
    real                 , intent(in)  :: h
    real   , dimension(n), intent(in)  :: f
    real   , dimension(n), intent(out) :: fl, fr
    logical, dimension(n), intent(in)  :: s

! local variables
!
    integer         :: i, im1, ip1
    real            :: dfl, dfr
    real            :: th, et, f1, f2, xl, xi

! saved variables
!
    logical, save   :: first = .true.
    real   , save   :: rdx = 1.0d+00, rdx2 = 1.0d+00
!
!-------------------------------------------------------------------------------
!
    if (first) then

! prepare parameters
!
      rdx   = rad * h
      rdx2  = rdx * rdx

! change the first execution indicator
!
      first = .false.

    end if

! iterate over positions and interpolate states
!
    do i = 1, n

! prepare neighbour indices
!
      im1 = max(1, i - 1)

! update the left and right-side interpolation states
!
#ifdef SHOCK_DETECTION
      if (s(i)) then

        fl(i) = f(i)
        fr(i) = f(i)

      else
#endif /* SHOCK_DETECTION */

! prepare neighbour indices
!
        ip1 = min(n, i + 1)

! prepare left and right differences
!
        dfr = f(ip1) - f(i  )
        dfl = f(i  ) - f(im1)

! calculate the indicator function (eq. 3.17 in [1])
!
        et = (dfl * dfl + dfr * dfr) / rdx2

! the switching function (embedded in eq. 3.22 in [1], eq. 32 in [2])
!
        xi = max(0.0d+00,                                                      &
                       0.5d+00 * min(2.0d+00, 1.0d+00 + (et - 1.0d+00) / eps))
        xl = 1.0d+00 - xi

! calculate values at i + ½
!
        if (dfr == 0.0d+00) then

          fl(i) = f(i)

        else

! calculate the slope ratio (eq. 2.8 in [1])
!
          th = dfl / dfr

! calculate the quadratic reconstruction (eq. 3.8 in [1], divided by 2)
!
          f1 = (2.0d+00 + th) / 6.0d+00

! calculate the third order limiter (eq. 3.13 in [1], cofficients divided by 2)
!
          if (th >= 0.0d+00) then
            f2 = max(0.0d+00, min(f1, th, 0.8d+00))
          else
            f2 = max(0.0d+00, min(f1, - 0.25d+00 * th))
          end if

! interpolate the left state (eq. 3.5 in [1], eq. 30 in [2])
!
          fl(i) = f(i) + dfr * (xl * f1 + xi * f2)

        end if

! calculate values at i - ½
!
        if (dfl == 0.0d+00) then

          fr(i) = f(i)

        else

! calculate the slope ratio (eq. 2.8 in [1])
!
          th = dfr / dfl

! calculate the quadratic reconstruction (eq. 3.8 in [1], divided by 2)
!
          f1 = (2.0d+00 + th) / 6.0d+00

! calculate the third order limiter (eq. 3.13 in [1], cofficients divided by 2)
!
          if (th >= 0.0d+00) then
            f2 = max(0.0d+00, min(f1, th, 0.8d+00))
          else
            f2 = max(0.0d+00, min(f1, - 0.25d+00 * th))
          end if

! interpolate the right state (eq. 3.5 in [1], eq. 30 in [2])
!
          fr(i) = f(i) - dfl * (xl * f1 + xi * f2)

        end if

#ifdef SHOCK_DETECTION
      end if ! shock vicinity
#endif /* SHOCK_DETECTION */

! shift the right state
!
      fr(im1) = fr(i)

    end do

! update the interpolation of the first and last points
!
    fl(1) = f (1)
    fr(n) = fl(n)

!-------------------------------------------------------------------------------
!
  end subroutine limo3
!
#endif /* LIMO3 */
#ifdef WENO3
!
!===============================================================================
!
! subroutine WENO3:
! ----------------
!
!   Subroutine reconstructs the interface states using the third order
!   Weighted Essentially Non-Oscillatory (WENO) method.
!
!   Arguments are described in subroutine reconstruct().
!
!   References:
!
!     [1] Yamaleev & Carpenter, 2009, J. Comput. Phys., 228, 3025
!
!===============================================================================
!
  subroutine weno3(n, h, f, fl, fr, s)

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer              , intent(in)  :: n
    real                 , intent(in)  :: h
    real   , dimension(n), intent(in)  :: f
    real   , dimension(n), intent(out) :: fl, fr
    logical, dimension(n), intent(in)  :: s

! local variables
!
    integer          :: i, im1, ip1
    real             :: h2, bp, bm, ap, am, wp, wm, ww
    real             :: dfl, dfr, df, fp, fm, fc

! selection weights
!
    real, parameter  :: dp = 2.0d0 / 3.0d0, dm = 1.0d0 / 3.0d0
!
!-------------------------------------------------------------------------------
!
! prepare common parameters
!
    h2 = h * h

! iterate along the vector
!
    do i = 1, n

! prepare neighbour indices
!
      im1     = max(1, i - 1)

! update the left and right-side interpolation states
!
#ifdef SHOCK_DETECTION
      if (s(i)) then

        fl(i) = f(i)
        fr(i) = f(i)

      else
#endif /* SHOCK_DETECTION */

! prepare neighbour indices
!
        ip1   = min(n, i + 1)

! calculate the left and right derivatives
!
        dfl   = f(i  ) - f(im1)
        dfr   = f(ip1) - f(i  )

! calculate coefficient omega
!
        ww    = (dfr - dfl)**2

! calculate corresponding betas
!
        bp    = dfr * dfr
        bm    = dfl * dfl

! calculate improved alphas
!
        ap    = 1.0d0 + ww / (bp + h2)
        am    = 1.0d0 + ww / (bm + h2)

! calculate weights
!
        wp    = dp * ap
        wm    = dm * am
        ww    = 2.0d0 * (wp + wm)

! calculate central interpolation
!
        fp    =   f(i  ) +         f(ip1)

! calculate left side interpolation
!
        fm    = - f(im1) + 3.0d0 * f(i  )

! calculate the left state
!
        fl(i) = (wp * fp + wm * fm) / ww

! calculate weights
!
        wp    = dp * am
        wm    = dm * ap
        ww    = 2.0d0 * (wp + wm)

! calculate central interpolation
!
        fp    =   f(i  ) +         f(im1)

! calculate right side interpolation
!
        fm    = - f(ip1) + 3.0d0 * f(i  )

! calculate the right state
!
        fr(i) = (wp * fp + wm * fm) / ww

#ifdef SHOCK_DETECTION
      end if ! shock vicinity
#endif /* SHOCK_DETECTION */

! shift the right state
!
      fr(im1) = fr(i)

    end do

! update the interpolation of the first and last points
!
    fl(1) = f (1)
    fr(n) = fl(n)

!-------------------------------------------------------------------------------
!
  end subroutine weno3
#endif /* WENO3 */
#if defined WENO5 || defined WENOZ
!
!===============================================================================
!
! subroutine WENOZ:
! ----------------
!
!   Subroutine reconstructs the interface states using the fifth order
!   Weighted Essentially Non-Oscillatory (WENO) method.  There are two
!   implementations which differ only by the used smoothness indicators.
!
!   Arguments are described in subroutine reconstruct().
!
!   References:
!
!     [1] Borges et al., 2008, J. Comput. Phys., 227, 3191
!
!===============================================================================
!
  subroutine wenoz(n, h, f, fl, fr, s)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer              , intent(in)  :: n
    real                 , intent(in)  :: h
    real   , dimension(n), intent(in)  :: f
    real   , dimension(n), intent(out) :: fl, fr
    logical, dimension(n), intent(in)  :: s

! local variables
!
    integer            :: i, im1, ip1, im2, ip2
    real               :: bl, bc, br, al, ac, ar, ww, wl, wc, wr, df
    real, dimension(3) :: ql, qr
    real, dimension(n) :: dfm, dfp, df2

! weight coefficients
!
    real, parameter :: c1 = 13.0d0 / 12.0d0, c2 = 0.25d0
    real, parameter :: dl = 0.1d0, dc = 0.6d0, dr = 0.3d0

! interpolation coefficients
!
    real, parameter :: a11 =   1.0d0 / 3.0d0, a12 = - 7.0d0 / 6.0d0            &
                     , a13 =  11.0d0 / 6.0d0
    real, parameter :: a21 = - 1.0d0 / 6.0d0, a22 =   5.0d0 / 6.0d0            &
                     , a23 =   1.0d0 / 3.0d0
    real, parameter :: a31 =   1.0d0 / 3.0d0, a32 =   5.0d0 / 6.0d0            &
                     , a33 = - 1.0d0 / 6.0d0
!
!-------------------------------------------------------------------------------
!
! calculate the left and right derivatives
!
    do i = 1, n - 1
      ip1 = i + 1

      dfp(i  ) = f(ip1) - f(i)
      dfm(ip1) = dfp(i)
    end do
    dfm(1) = dfp(1)
    dfp(n) = dfm(n)
    df2(:) = c1 * (dfp(:) - dfm(:))**2

! iterate along the vector
!
    do i = 1, n

! prepare neighbour indices
!
      im1     = max(1, i - 1)

! update the left and right-side interpolation states
!
#ifdef SHOCK_DETECTION
      if (s(i)) then

        fl(i) = f(i)
        fr(i) = f(i)

      else
#endif /* SHOCK_DETECTION */

! prepare neighbour indices
!
        im2     = max(1, i - 2)
        ip1     = min(n, i + 1)
        ip2     = min(n, i + 2)

! calculate betas
!
        bl      = df2(im1) + c2 * (3.0d0 * dfm(i  ) - dfm(im1))**2
        bc      = df2(i  ) + c2 * (        dfp(i  ) + dfm(i  ))**2
        br      = df2(ip1) + c2 * (3.0d0 * dfp(i  ) - dfp(ip1))**2
#ifdef WENOZ
        ww      = abs(bl - br)
#endif /* WENOZ */

! calculate alphas
!
#ifdef WENO5
        al      = 1.0d0 / (bl + eps)**2
        ac      = 1.0d0 / (bc + eps)**2
        ar      = 1.0d0 / (br + eps)**2
#endif /* WENO5 */
#ifdef WENOZ
        al      = 1.0d0 + ww / (bl + eps)
        ac      = 1.0d0 + ww / (bc + eps)
        ar      = 1.0d0 + ww / (br + eps)
#endif /* WENOZ */

! calculate weights
!
        wl      = dl * al
        wc      = dc * ac
        wr      = dr * ar
        ww      = wl + wc + wr

! calculate the second order interpolations of the left state
!
        ql(1)   = a11 * f(im2) + a12 * f(im1) + a13 * f(i  )
        ql(2)   = a21 * f(im1) + a22 * f(i  ) + a23 * f(ip1)
        ql(3)   = a31 * f(i  ) + a32 * f(ip1) + a33 * f(ip2)

! calculate the left state
!
        fl(i)   = (wl * ql(1) + wc * ql(2) + wr * ql(3)) / ww

! calculate weights
!
        wl      = dl * ar
        wc      = dc * ac
        wr      = dr * al
        ww      = wl + wc + wr

! calculate the second order interpolations of the right state
!
        qr(1)   = a11 * f(ip2) + a12 * f(ip1) + a13 * f(i  )
        qr(2)   = a21 * f(ip1) + a22 * f(i  ) + a23 * f(im1)
        qr(3)   = a31 * f(i  ) + a32 * f(im1) + a33 * f(im2)

! calculate the right state
!
        fr(i) = (wl * qr(1) + wc * qr(2) + wr * qr(3)) / ww

#ifdef SHOCK_DETECTION
      end if ! shock vicinity
#endif /* SHOCK_DETECTION */

! shift the right state
!
      fr(im1) = fr(i)

    end do

! update the interpolation of the first and last points
!
    fl(1) = f (1)
    fr(n) = fl(n)

!-------------------------------------------------------------------------------
!
  end subroutine wenoz
#endif /* WENO5 | WENOZ */
#ifdef MP
!
!===============================================================================
!
! subroutine MP:
! -------------
!
!   Subroutine reconstructs the interface states using the third and higher
!   order Monotonicity Preserving (MP) method.  This is an improved version
!   by He et al..
!
!   Arguments are described in subroutine reconstruct().
!
!   References:
!
!     [1] Suresh, A. & Huynh, H. T., "Accurate Monotonicity-Preserving Schemes
!         with Runge-Kutta Time Stepping",
!         Journal of Computational Physics, 1997, 136, pp. 83-99
!     [2] He, Z.-W., Li, X.-L., FU, D.-X., & Ma, Y.-W., "A 5th order
!         monotonicity-preserving upwind compact difference scheme",
!         Science China: Physics, Mechanics & Astronomy, 2011, 54, pp. 511-522
!
!===============================================================================
!
  subroutine mp(n, h, f, fl, fr, s)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer              , intent(in)  :: n
    real                 , intent(in)  :: h
    real   , dimension(n), intent(in)  :: f
    real   , dimension(n), intent(out) :: fl, fr
    logical, dimension(n), intent(in)  :: s

! local variables
!
    logical            :: flag
    integer            :: i
    integer            :: im1, ip1
#if MP_ORDER > 3
    integer            :: im2, ip2
#endif /* MP_ORDER > 3 */
#if MP_ORDER > 5
    integer            :: im3, ip3
#endif /* MP_ORDER > 5 */
#if MP_ORDER > 7
    integer            :: im4, ip4
#endif /* MP_ORDER > 7 */
    real               :: df, fmp, fmd, ful, flc, fmn, fmx
    real               :: ds, dm1, dc0, dp1, dc4, dml, dmr
    real               :: sigma
    real, dimension(n) :: fp, fm, dfm, dfp

! parameters
!
    real, parameter    :: a31 = -   1.0d0 /    6.0d0                           &
                        , a32 =     5.0d0 /    6.0d0                           &
                        , a33 =     1.0d0 /    3.0d0
    real, parameter    :: a51 =     1.0d0 /   30.0d0                           &
                        , a52 = -  13.0d0 /   60.0d0                           &
                        , a53 =    47.0d0 /   60.0d0                           &
                        , a54 =     9.0d0 /   20.0d0                           &
                        , a55 = -   1.0d0 /   20.0d0
    real, parameter    :: a71 = -   1.0d0 /  140.0d0                           &
                        , a72 =     5.0d0 /   84.0d0                           &
                        , a73 = - 101.0d0 /  420.0d0                           &
                        , a74 =   319.0d0 /  420.0d0                           &
                        , a75 =   107.0d0 /  210.0d0                           &
                        , a76 = -  19.0d0 /  210.0d0                           &
                        , a77 =     1.0d0 /  105.0d0
    real, parameter    :: a91 =     1.0d0 /  630.0d0                           &
                        , a92 = -  41.0d0 / 2520.0d0                           &
                        , a93 =   199.0d0 / 2520.0d0                           &
                        , a94 = - 641.0d0 / 2520.0d0                           &
                        , a95 =  1879.0d0 / 2520.0d0                           &
                        , a96 =   275.0d0 /  504.0d0                           &
                        , a97 = -  61.0d0 /  504.0d0                           &
                        , a98 =    11.0d0 /  504.0d0                           &
                        , a99 = -   1.0d0 /  504.0d0
!
!-------------------------------------------------------------------------------
!
! calculate the left and right derivatives
!
    do i = 1, n - 1
      ip1 = i + 1

      dfp(i  ) = f(ip1) - f(i)
      dfm(ip1) = dfp(i)
    end do
    dfm(1) = 0.0d0
    dfp(n) = 0.0d0

! iterate along the vector
!
    do i = 1, n

! prepare neighbour indices
!
      im1   = max(1, i - 1)
      ip1   = min(n, i + 1)

! update the left and right-side interpolation states
!
#ifdef SHOCK_DETECTION
      if (s(i)) then

        fl(i) = f(i)
        fr(i) = f(i)

      else

! calculate the polynomial interpolation
!
        if (i <= 1 .or. i >= n) then
          fp(i) = f(i)
          fm(i) = f(i)
        else
#if MP_ORDER > 3
          im2  = i - 2
          ip2  = i + 2
          if (im2 < 1 .or. ip2 > n .or. s(im1) .or. s(ip1)) then
            fp(i) = a31 * f(im1) + a32 * f(i  ) + a33 * f(ip1)
            fm(i) = a31 * f(ip1) + a32 * f(i  ) + a33 * f(im1)
          else
#if MP_ORDER > 5
            im3 = i - 3
            ip3 = i + 3
            if (im3 < 1 .or. ip3 > n .or. s(im2) .or. s(ip2)) then
              fp(i) = a51 * f(im2) + a52 * f(im1) + a53 * f(i  )               &
                    + a54 * f(ip1) + a55 * f(ip2)
              fm(i) = a51 * f(ip2) + a52 * f(ip1) + a53 * f(i  )               &
                    + a54 * f(im1) + a55 * f(im2)
            else
#if MP_ORDER > 7
              im4 = i - 4
              ip4 = i + 4
              if (im4 < 1 .or. ip4 > n .or. s(im3) .or. s(ip3)) then
                fp(i) = a71 * f(im3) + a72 * f(im2) + a73 * f(im1)             &
                      + a74 * f(i  ) + a75 * f(ip1) + a76 * f(ip2)             &
                      + a77 * f(ip3)
                fm(i) = a71 * f(ip3) + a72 * f(ip2) + a73 * f(ip1)             &
                      + a74 * f(i  ) + a75 * f(im1) + a76 * f(im2)             &
                      + a77 * f(im3)
              else
                fp(i) = a91 * f(im4) + a92 * f(im3) + a93 * f(im2)             &
                      + a94 * f(im1) + a95 * f(i  ) + a96 * f(ip1)             &
                      + a97 * f(ip2) + a98 * f(ip3) + a99 * f(ip4)
                fm(i) = a91 * f(ip4) + a92 * f(ip3) + a93 * f(ip2)             &
                      + a94 * f(ip1) + a95 * f(i  ) + a96 * f(im1)             &
                      + a97 * f(im2) + a98 * f(im3) + a99 * f(im4)
              end if
#else /* MP_ORDER > 7 */
              fp(i) = a71 * f(im3) + a72 * f(im2) + a73 * f(im1)               &
                    + a74 * f(i  ) + a75 * f(ip1) + a76 * f(ip2) + a77 * f(ip3)
              fm(i) = a71 * f(ip3) + a72 * f(ip2) + a73 * f(ip1)               &
                    + a74 * f(i  ) + a75 * f(im1) + a76 * f(im2) + a77 * f(im3)
#endif /* MP_ORDER > 7 */
            end if
#else /* MP_ORDER > 5 */
            fp(i) = a51 * f(im2) + a52 * f(im1) + a53 * f(i  ) + a54 * f(ip1)  &
                  + a55 * f(ip2)
            fm(i) = a51 * f(ip2) + a52 * f(ip1) + a53 * f(i  ) + a54 * f(im1)  &
                  + a55 * f(im2)
#endif /* MP_ORDER > 5 */
          end if
#else /* MP_ORDER > 3 */
          fp(i) = a31 * f(im1) + a32 * f(i  ) + a33 * f(ip1)
          fm(i) = a31 * f(ip1) + a32 * f(i  ) + a33 * f(im1)
#endif /* MP_ORDER > 3 */
        end if

! prepare sigma
!
        if (dfm(i) * dfp(i) >= 0.0d+00) then
          sigma = kappa
        else
          sigma = kbeta
        end if

! calculate the monotonicity-preserving limit for the interpolation
!
        df    = sigma * dfm(i)
        fmp   = f(i) + minmod(dfp(i), df)

! reset the flag indicating that the second order derivatives have to be
! calculated
!
        flag    = .true.

! calculate the monotonicity-preserving condition for the left-side interface
! (eq. 2.30 in [1])
!
        ds    = (fp(i) - f(i)) * (fp(i) - fmp)

! if the condition holds, return the value of the left-side interface,
! otherwise apply limiting procedure
!
        if (ds <= eps) then

          fl(i) = fp(i)

        else

! calculate the limited second order derivatives
!
          dm1   = dfp(im1) - dfm(im1)
          dc0   = dfp(i  ) - dfm(i  )
          dp1   = dfp(ip1) - dfm(ip1)
          dc4   = 4.0d0 * dc0

          dml   = 0.5d0 * minmod4(dc4 - dm1, 4.0d0 * dm1 - dc0, dc0, dm1)
          dmr   = 0.5d0 * minmod4(dc4 - dp1, 4.0d0 * dp1 - dc0, dc0, dp1)

          flag  = .false.

! calculate limiting values from eqs. (14), (18a), and (20b') in [2]
!
          fmd   = f(i) + 0.5d0 * dfp(i) - dmr
          ful   = f(i) +         df
          flc   = f(i) + 0.5d0 * df     + dml

! estimate the limits
!
          fmx   = max(min(f(i), f(ip1), fmd), min(f(i), ful, flc))
          fmn   = min(max(f(i), f(ip1), fmd), max(f(i), ful, flc))

! limit the left state
!
          fl(i) = median(fp(i), fmn, fmx)

        end if

! calculate the upper limit for interpolation
!
        df    = sigma * dfp(i)
        fmp   = f(i) - minmod(dfm(i), df)

! calculate the monotonicity-preserving condition for the right-side interface
! (eq. 2.30 in [1])
!
        ds    = (fm(i) - f(i)) * (fm(i) - fmp)

! if the condition holds, return the value of the left-side interface,
! otherwise apply limiting procedure
!
        if (ds <= eps) then

          fr(i) = fm(i)

        else

! check if we have to calculate the limited second order derivatives
!
          if (flag) then

! calculate the limited second order derivatives
!
            dm1 = dfp(im1) - dfm(im1)
            dc0 = dfp(i  ) - dfm(i  )
            dp1 = dfp(ip1) - dfm(ip1)
            dc4 = 4.0d0 * dc0

            dml = 0.5d0 * minmod4(dc4 - dm1, 4.0d0 * dm1 - dc0, dc0, dm1)
            dmr = 0.5d0 * minmod4(dc4 - dp1, 4.0d0 * dp1 - dc0, dc0, dp1)

          end if

! calculate limiting values from eqs. (14), (18a), and (20b') in [2]
!
          fmd   = f(i) - 0.5d0 * dfm(i) - dml
          ful   = f(i) -         df
          flc   = f(i) - 0.5d0 * df     + dmr

! estimate the limits
!
          fmx   = max(min(f(i), f(im1), fmd), min(f(i), ful, flc))
          fmn   = min(max(f(i), f(im1), fmd), max(f(i), ful, flc))

! limit the right state
!
          fr(i) = median(fm(i), fmn, fmx)

        end if

      end if ! shock vicinity
#else /* SHOCK_DETECTION */

! calculate the polynomial interpolation
!
      if (i <= 1 .or. i >= n) then
        fp(i) = f(i)
        fm(i) = f(i)
      else
#if MP_ORDER > 3
        im2 = i - 2
        ip2 = i + 2
        if (im2 < 1 .or. ip2 > n) then
          fp(i) = a31 * f(im1) + a32 * f(i  ) + a33 * f(ip1)
          fm(i) = a31 * f(ip1) + a32 * f(i  ) + a33 * f(im1)
        else
#if MP_ORDER > 5
          im3 = i - 3
          ip3 = i + 3
          if (im3 < 1 .or. ip3 > n) then
            fp(i) = a51 * f(im2) + a52 * f(im1) + a53 * f(i  ) + a54 * f(ip1)  &
                  + a55 * f(ip2)
            fm(i) = a51 * f(ip2) + a52 * f(ip1) + a53 * f(i  ) + a54 * f(im1)  &
                  + a55 * f(im2)
          else
#if MP_ORDER > 7
            im4 = i - 4
            ip4 = i + 4
            if (im4 < 1 .or. ip4 > n) then
              fp(i) = a71 * f(im3) + a72 * f(im2) + a73 * f(im1)               &
                    + a74 * f(i  ) + a75 * f(ip1) + a76 * f(ip2)               &
                    + a77 * f(ip3)
              fm(i) = a71 * f(ip3) + a72 * f(ip2) + a73 * f(ip1)               &
                    + a74 * f(i  ) + a75 * f(im1) + a76 * f(im2)               &
                    + a77 * f(im3)
            else
              fp(i) = a91 * f(im4) + a92 * f(im3) + a93 * f(im2)               &
                    + a94 * f(im1) + a95 * f(i  ) + a96 * f(ip1)               &
                    + a97 * f(ip2) + a98 * f(ip3) + a99 * f(ip4)
              fm(i) = a91 * f(ip4) + a92 * f(ip3) + a93 * f(ip2)               &
                    + a94 * f(ip1) + a95 * f(i  ) + a96 * f(im1)               &
                    + a97 * f(im2) + a98 * f(im3) + a99 * f(im4)
            end if
#else /* MP_ORDER > 7 */
            fp(i) = a71 * f(im3) + a72 * f(im2) + a73 * f(im1) + a74 * f(i  )  &
                  + a75 * f(ip1) + a76 * f(ip2) + a77 * f(ip3)
            fm(i) = a71 * f(ip3) + a72 * f(ip2) + a73 * f(ip1) + a74 * f(i  )  &
                  + a75 * f(im1) + a76 * f(im2) + a77 * f(im3)
#endif /* MP_ORDER > 7 */
          end if
#else /* MP_ORDER > 5 */
          fp(i) = a51 * f(im2) + a52 * f(im1) + a53 * f(i  ) + a54 * f(ip1)    &
                + a55 * f(ip2)
          fm(i) = a51 * f(ip2) + a52 * f(ip1) + a53 * f(i  ) + a54 * f(im1)    &
                + a55 * f(im2)
#endif /* MP_ORDER > 5 */
        end if
#else /* MP_ORDER > 3 */
        fp(i) = a31 * f(im1) + a32 * f(i  ) + a33 * f(ip1)
        fm(i) = a31 * f(ip1) + a32 * f(i  ) + a33 * f(im1)
#endif /* MP_ORDER > 3 */
      end if

! prepare sigma
!
      if (dfm(i) * dfp(i) >= 0.0d+00) then
        sigma = kappa
      else
        sigma = kbeta
      end if

! calculate the monotonicity-preserving limit for the interpolation
!
      df    = sigma * dfm(i)
      fmp   = f(i) + minmod(dfp(i), df)

! reset the flag indicating that the second order derivatives have to be
! calculated
!
      flag    = .true.

! calculate the monotonicity-preserving condition for the left-side interface
! (eq. 2.30 in [1])
!
      ds    = (fp(i) - f(i)) * (fp(i) - fmp)

! if the condition holds, return the value of the left-side interface,
! otherwise apply limiting procedure
!
      if (ds <= eps) then

        fl(i) = fp(i)

      else

! calculate the limited second order derivatives
!
        dm1   = dfp(im1) - dfm(im1)
        dc0   = dfp(i  ) - dfm(i  )
        dp1   = dfp(ip1) - dfm(ip1)
        dc4   = 4.0d0 * dc0

        dml   = 0.5d0 * minmod4(dc4 - dm1, 4.0d0 * dm1 - dc0, dc0, dm1)
        dmr   = 0.5d0 * minmod4(dc4 - dp1, 4.0d0 * dp1 - dc0, dc0, dp1)

        flag  = .false.

! calculate limiting values from eqs. (14), (18a), and (20b') in [2]
!
        fmd   = f(i) + 0.5d0 * dfp(i) - dmr
        ful   = f(i) +         df
        flc   = f(i) + 0.5d0 * df     + dml

! estimate the limits
!
        fmx   = max(min(f(i), f(ip1), fmd), min(f(i), ful, flc))
        fmn   = min(max(f(i), f(ip1), fmd), max(f(i), ful, flc))

! limit the left state
!
        fl(i) = median(fp(i), fmn, fmx)

      end if

! calculate the upper limit for interpolation
!
      df    = sigma * dfp(i)
      fmp   = f(i) - minmod(dfm(i), df)

! calculate the monotonicity-preserving condition for the right-side interface
! (eq. 2.30 in [1])
!
      ds    = (fm(i) - f(i)) * (fm(i) - fmp)

! if the condition holds, return the value of the left-side interface,
! otherwise apply limiting procedure
!
      if (ds <= eps) then

        fr(i) = fm(i)

      else

! check if we have to calculate the limited second order derivatives
!
        if (flag) then

! calculate the limited second order derivatives
!
          dm1 = dfp(im1) - dfm(im1)
          dc0 = dfp(i  ) - dfm(i  )
          dp1 = dfp(ip1) - dfm(ip1)
          dc4 = 4.0d0 * dc0

          dml = 0.5d0 * minmod4(dc4 - dm1, 4.0d0 * dm1 - dc0, dc0, dm1)
          dmr = 0.5d0 * minmod4(dc4 - dp1, 4.0d0 * dp1 - dc0, dc0, dp1)

        end if

! calculate limiting values from eqs. (14), (18a), and (20b') in [2]
!
        fmd   = f(i) - 0.5d0 * dfm(i) - dml
        ful   = f(i) -         df
        flc   = f(i) - 0.5d0 * df     + dmr

! estimate the limits
!
        fmx   = max(min(f(i), f(im1), fmd), min(f(i), ful, flc))
        fmn   = min(max(f(i), f(im1), fmd), max(f(i), ful, flc))

! limit the right state
!
        fr(i) = median(fm(i), fmn, fmx)

      end if
#endif /* SHOCK_DETECTION */

! shift the right state
!
      fr(im1) = fr(i)

    end do

! update the interpolation of the first and last points
!
    fl(1) = fr(1)
    fr(n) = fl(n)

!-------------------------------------------------------------------------------
!
  end subroutine mp
#endif /* MP */
#if defined TVD || defined CENO3
!
!===============================================================================
!
! subroutine LIMITER:
!
!   Subroutine calculates the local derivative applying a selected TVD limiter.
!   Two limiters are implemented, MINMOD and LF.
!
!===============================================================================
!
  subroutine limiter(n, f, df)

    implicit none

! input/output arguments
!
    integer           , intent(in)  :: n
    real, dimension(n), intent(in)  :: f
    real, dimension(n), intent(out) :: df

! local variables
!
    integer            :: i, ip1
    real               :: ds
    real, dimension(n) :: dfl, dfr
!
!------------------------------------------------------------------------------
!
! calculate the left- and right-side derivatives
!
    dfl(1) = 0.0
    do i = 1, n - 1
      ip1 = i + 1

      dfr(i  ) = f(ip1) - f(i)
      dfl(ip1) = dfr(i)
    end do
    dfr(n) = 0.0

! second order interpolation
!
    do i = 1, n

#ifdef MMLIM
      df(i) = minmod(dfl(i), dfr(i))
#endif /* MMLIM */
#ifdef MCLIM
      df(i) = (sign(0.5, dfr(i)) + sign(0.5, dfl(i)))                          &
                                  * min(2.0 * abs(dfr(i))                      &
                                      , 0.5 * abs(dfr(i) + dfl(i))             &
                                      , 2.0 * abs(dfl(i)))
#endif /* MCLIM */
#ifdef LFLIM
! calculate the monotonicity indicator
!
      ds = dfr(i) * dfl(i)

! if the region is monotonic calculate limited derivative, otherwise set zero
!
      if (ds .gt. eps) then

! use selected limiter
!
        df(i) = 2.0 * ds / (dfr(i) + dfl(i))
      else
        df(i) = 0.0
      end if
#endif /* LFLIM */

    end do

!-------------------------------------------------------------------------------
!
  end subroutine limiter
#endif /* TVD | CENO3 */
!
!===============================================================================
!
! function MINMOD:
! ===============
!
!   Function returns the minimum module value among two arguments.
!
!   Arguments:
!
!     a, b - the input values;
!
!===============================================================================
!
  real function minmod(a, b)

! local variables are not implicit by default
!
    implicit none

! input arguments
!
    real, intent(in) :: a, b
!
!-------------------------------------------------------------------------------
!
    minmod = (sign(0.5d+00, a) + sign(0.5d+00, b)) * min(abs(a), abs(b))

    return

!-------------------------------------------------------------------------------
!
  end function minmod
!
!===============================================================================
!
! function MINMOD4:
! ================
!
!   Function returns the minimum module value among four arguments.
!
!   Arguments:
!
!     a, b, c, d - the input values;
!
!===============================================================================
!
  real function minmod4(a, b, c, d)

! local variables are not implicit by default
!
    implicit none

! input arguments
!
    real, intent(in) :: a, b, c, d
!
!-------------------------------------------------------------------------------
!
    minmod4 = minmod(minmod(a, b), minmod(c, d))

    return

!-------------------------------------------------------------------------------
!
  end function minmod4
!
!===============================================================================
!
! function MEDIAN:
! ===============
!
!   Function returns the median of three argument values.
!
!   Arguments:
!
!     a, b, c - the input values;
!
!===============================================================================
!
  real function median(a, b, c)

! local variables are not implicit by default
!
    implicit none

! input arguments
!
    real, intent(in) :: a, b, c
!
!-------------------------------------------------------------------------------
!
    median = a + minmod(b - a, c - a)

    return

  end function median
!
!===============================================================================
!
! correct_reconstruction: subroutine checks the reconstruction in order to
!                         improve the stability
!
!===============================================================================
!
  subroutine correct_reconstruction(n, f, fl, fr)

! input/output arguments
!
    integer           , intent(in)    :: n
    real, dimension(n), intent(in)    :: f
    real, dimension(n), intent(inout) :: fl, fr

! local variables
!
    integer               :: i, im1, ip1
    real                  :: fmn, fmx
!
!------------------------------------------------------------------------------
!
! iterate over all points
!
    do i = 1, n

! calculate indices
!
      im1 = max(1, i - 1)
      ip1 = min(n, i + 1)

! estimate the bounds of the allowed interval for reconstructed states
!
      fmn = min(f(i), f(ip1))
      fmx = max(f(i), f(ip1))

! check if the left state lays in the allowed range
!
      if (fl(i) < fmn .or. fl(i) > fmx) then

! calculate new states
!
        fl(i  ) = f(i  )
        fr(im1) = f(i  )

      end if

! check if the right state lays in the allowed range
!
      if (fr(i) < fmn .or. fr(i) > fmx) then

! calculate new states
!
        fl(ip1) = f(ip1)
        fr(i  ) = f(ip1)

      end if

    end do

!-------------------------------------------------------------------------------
!
  end subroutine correct_reconstruction
!
!===============================================================================
!
! subroutine FIX_POSITIVITY:
! -------------------------
!
!   Subroutine scans the input arrays of the left and right states fl(:) and
!   fr(:) for negative values.  If it finds a negative value, it repeates the
!   state reconstruction from f(:) using the zeroth order interpolation.
!
!===============================================================================
!
  subroutine fix_positivity(n, f, fl, fr)

! local variables are not implicit by default
!
    implicit none

! input/output arguments
!
    integer           , intent(in)    :: n
    real, dimension(n), intent(in)    :: f
    real, dimension(n), intent(inout) :: fl, fr

! local variables
!
    integer :: i, im1, ip1
    real    :: fmn, fmx
!
!------------------------------------------------------------------------------
!
! look for negative values in the states along the vector
!
    do i = 1, n

! check if there is a negative value in the left or right states
!
      if (min(fl(i), fr(i)) .le. 0.0d0) then

! calculate the left and right indices
!
        im1 = max(1, i - 1)
        ip1 = min(n, i + 1)

! prepare the allowed interval
!
        fmn = min(f(i), f(ip1))
        fmx = max(f(i), f(ip1))

! limit the states using the zeroth-order reconstruction
!
        if ((fl(i) .lt. fmn) .or. (fl(i) .gt. fmx)) then
          fl(i  ) = f(i)
          fr(im1) = f(i)
        end if

        if ((fr(i) .lt. fmn) .or. (fr(i) .gt. fmx)) then
          fl(ip1) = f(ip1)
          fr(i  ) = f(ip1)
        end if
      end if
    end do

!-------------------------------------------------------------------------------
!
  end subroutine fix_positivity

!===============================================================================
!
end module interpolations
