!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: PROBLEMS
!!
!!  This module handles the initialization of various test and research
!!  problems.
!!
!!******************************************************************************
!
module problems

! include external procedures and variables
!
  use parameters, only : get_parameter_integer, get_parameter_real             &
                       , get_parameter_string

! module variables are not implicit by default
!
  implicit none

! module variable to store the problem name
!
  character(len=32), save :: problem = "blast"

! by default everything is private
!
  private

! declare public subroutines
!
  public :: initialize_problem

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! subroutine INITIALIZE_PROBLEM:
! -----------------------------
!
!   Subroutine call a subroutine to set the initial conditions for selected
!   problem.
!
!===============================================================================
!
  subroutine initialize_problem(iret)

! local variables are not implicit by default
!
    implicit none

! subroutine arguments
!
    integer, intent(out) :: iret
!
!-------------------------------------------------------------------------------
!
! get the problem name from module PARAMETERS
!
    call get_parameter_string("problem", problem)

! reset the output flag
!
    iret = 0

! select the initialization subroutine depending on the problem name
!
    select case(problem)

! general problems, useful to test all sets of equations
!
    case("riemann")
      call riemann()
    case("blast")
      call blast()
    case("rotor")
      call rotor()
    case("implosion")
      call implosion()
    case("kh", "kelvinhelmholtz", "kelvin-helmholtz")
      call kh()
    case("rt", "rayleightaylor", "rayleigh-taylor")
      call rt()
    case("turbulence")
      call turbulence()

#ifdef MHD
! magnetohydrodynamic problems
!
    case("ot", "orszagtang", "orszag-tang")
      call ot()
    case("loop")
      call loop()
#endif /* MHD */

    case default
      iret = 101
    end select

!-------------------------------------------------------------------------------
!
  end subroutine initialize_problem
!
!===============================================================================
!
! subroutine RIEMANN:
! ------------------
!
!   Subroutine sets up the initial conditions for general Riemann problem.
!
!===============================================================================
!
  subroutine riemann()

! include external procedures and variables
!
    use equations, only : prim2cons
#ifdef MHD
    use equations, only : update_magnetic_variables
#endif /* MHD */
    use mesh     , only : im, jm, km, x, xi
    use variables, only : nu, nt
    use variables, only : uu, qp, qp_bnd
#ifdef MHD
    use variables, only : bext
#endif /* MHD */
    use variables, only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables, only : ipr
#endif /* ADI */
#ifdef MHD
    use variables, only : inx, iny, inz
    use variables, only : ibx, iby, ibz
#ifdef VPOT
    use variables, only : iax, iay, iaz
#endif /* VPOT */
#endif /* MHD */

! local variables are not implicit by default
!
    implicit none

! local variables
!
    integer :: i, j, k, jc, kc

! temporary array to store primitive variables
!
    real, dimension(nt,im) :: q, u
!
!------------------------------------------------------------------------------
!
! prepare the indices of the center
!
    jc = max(1, jm / 2)
    kc = max(1, km / 2)

#ifdef MHD
! check requrements
!
    if (qp_bnd(ibx,1,1) == qp_bnd(ibx,1,2)) then
      bext(inx) = qp_bnd(ibx,1,1)
    else
      write(*,*) "ERROR: Bx cannot be discontinuous in the Riemann problem!"
      stop
    end if

#ifdef GLM
#else /* GLM */
! set the initial vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    uu(iaz,:,:,:) = 0.0d0

    if (qp_bnd(iby,1,1) == qp_bnd(iby,1,2)) then
      bext(iny) = qp_bnd(iby,1,1)
    else
      bext(iny) = 0.0d0
#ifdef CT
      do i = 1, im
        if (xi(i) < 0.0d0) then
          uu(iaz,i,:,:) = - qp_bnd(iby,1,1) * xi(i)
        else if (xi(i) > 0.0d0) then
          uu(iaz,i,:,:) = - qp_bnd(iby,1,2) * xi(i)
        else
          uu(iaz,i,:,:) = 0.0d0
        end if
      end do
#endif /* CT */
#ifdef CD
      do i = 1, im
        if (x(i) < 0.0d0) then
          uu(iaz,i,:,:) = - qp_bnd(iby,1,1) * x(i)
        else if (x(i) > 0.0d0) then
          uu(iaz,i,:,:) = - qp_bnd(iby,1,2) * x(i)
        else
          uu(iaz,i,:,:) = 0.0d0
        end if
      end do
#endif /* CD */
    end if

#ifndef R2D
    if (qp_bnd(ibz,1,1) == qp_bnd(ibz,1,2)) then
      bext(inz) = qp_bnd(ibz,1,1)
    else
      bext(inz) = 0.0d0
#ifdef CT
      do i = 1, im
        if (xi(i) < 0.0d0) then
          uu(iay,i,:,:) = qp_bnd(ibz,1,1) * xi(i)
        else if (xi(i) > 0.0d0) then
          uu(iay,i,:,:) = qp_bnd(ibz,1,2) * xi(i)
        else
          uu(iay,i,:,:) = 0.0d0
        end if
      end do
#endif /* CT */
#ifdef CD
      do i = 1, im
        if (x(i) < 0.0d0) then
          uu(iay,i,:,:) = qp_bnd(ibz,1,1) * x(i)
        else if (x(i) > 0.0d0) then
          uu(iay,i,:,:) = qp_bnd(ibz,1,2) * x(i)
        else
          uu(iay,i,:,:) = 0.0d0
        end if
      end do
#endif /* CD */
    end if
#endif /* !R2D */
#endif /* GLM */

! preprare magnetic variables
!
    call update_magnetic_variables()
#endif /* MHD */

! set the problem variables
!
    do i = 1, im

      if (x(i) < 0.0d0) then

! set the variables of the left state
!
        q(idn,i) = qp_bnd(idn,1,1)
        q(ivx,i) = qp_bnd(ivx,1,1)
        q(ivy,i) = qp_bnd(ivy,1,1)
        q(ivz,i) = qp_bnd(ivz,1,1)
#ifdef ADI
        q(ipr,i) = qp_bnd(ipr,1,1)
#endif /* ADI */
#ifdef MHD
        q(ibx,i) = qp(ibx,i,jc,kc)
        q(iby,i) = qp(iby,i,jc,kc)
        q(ibz,i) = qp(ibz,i,jc,kc)
#endif /* MHD */

      else if (x(i) > 0.0d0) then

! set the variables of the right state
!
        q(idn,i) = qp_bnd(idn,1,2)
        q(ivx,i) = qp_bnd(ivx,1,2)
        q(ivy,i) = qp_bnd(ivy,1,2)
        q(ivz,i) = qp_bnd(ivz,1,2)
#ifdef ADI
        q(ipr,i) = qp_bnd(ipr,1,2)
#endif /* ADI */
#ifdef MHD
        q(ibx,i) = qp(ibx,i,jc,kc)
        q(iby,i) = qp(iby,i,jc,kc)
        q(ibz,i) = qp(ibz,i,jc,kc)
#endif /* MHD */

      else

! set the variables of the intermediate state
!
        q(idn,i) = 0.5d0 * (qp_bnd(idn,1,1) + qp_bnd(idn,1,2))
        q(ivx,i) = 0.5d0 * (qp_bnd(ivx,1,1) + qp_bnd(ivx,1,2))
        q(ivy,i) = 0.5d0 * (qp_bnd(ivy,1,1) + qp_bnd(ivy,1,2))
        q(ivz,i) = 0.5d0 * (qp_bnd(ivz,1,1) + qp_bnd(ivz,1,2))
#ifdef ADI
        q(ipr,i) = 0.5d0 * (qp_bnd(ipr,1,1) + qp_bnd(ipr,1,2))
#endif /* ADI */
#ifdef MHD
        q(ibx,i) = qp(ibx,i,jc,kc)
        q(iby,i) = qp(iby,i,jc,kc)
        q(ibz,i) = qp(ibz,i,jc,kc)
#endif /* MHD */

      end if

    end do

! convert the primitive variables to conservative ones
!
    call prim2cons(im, q(1:nt,1:im), u(1:nt,1:im))

    do k = 1, km
      do j = 1, jm

! update array of primitive variables
!
        qp(1:nt,1:im,j,k) = q(1:nt,1:im)

! update array of conserved variables
!
        uu(1:nu,1:im,j,k) = u(1:nu,1:im)

      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine riemann
!
!===============================================================================
!
! subroutine BLAST:
! ----------------
!
!   Subroutine sets up the initial conditions for spherical blast wave.
!
!===============================================================================
!
  subroutine blast()

! include external procedures and variables
!
#ifdef ADI
    use equations, only : gamma
#endif /* ADI */
    use equations, only : prim2cons
#ifdef MHD
    use equations, only : update_magnetic_variables
#endif /* MHD */
    use mesh     , only : im, jm, km, x, y, z, dx, dy, dxh, dyh, dzh, daxy
    use variables, only : nu, nt
    use variables, only : uu, qp
#ifdef MHD
    use variables, only : bext
#endif /* MHD */
    use variables, only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables, only : ipr
#endif /* ADI */
#ifdef MHD
#ifdef GLM
    use variables, only : iph
#endif /* GLM */
    use variables, only : inx, iny, inz
    use variables, only : ibx, iby, ibz
#ifdef VPOT
    use variables, only : iax, iay, iaz
#endif /* VPOT */
#endif /* MHD */

! local variables are not implicit by default
!
    implicit none

! default parameter values
!
    real    :: dens = 1.0d0, ratio = 1.0e2, radius = 0.1d0
#ifdef ADI
    real    :: csnd = 0.40824829046386301635d0
#endif /* ADI */

! local variables
!
    integer :: i, j, k
    real    :: rad
    real    :: xl, yl, zl, xu, yu, zu, rl, ru
    real    :: xb, yb, xt, yt
    real    :: fc_amb, fc_ovr
    real    :: dn_amb, dn_ovr
#ifdef ADI
    real    :: pr_amb, pr_ovr
#endif /* ADI */

! temporary array to store primitive variables
!
    real, dimension(nt,im) :: q, u
!
!------------------------------------------------------------------------------
!
! get problem parameters from the parameter file
!
    call get_parameter_real("dens"  , dens  )
    call get_parameter_real("ratio" , ratio )
    call get_parameter_real("radius", radius)
#ifdef ADI
    call get_parameter_real("csnd"  , csnd  )
#endif /* ADI */

! calculate the overdense and ambient region densities
!
    dn_amb = dens
#ifdef ISO
    dn_ovr = dn_amb * ratio
#endif /* ISO */
#ifdef ADI
    dn_ovr = dn_amb
#endif /* ADI */

#ifdef ADI
! calculate parallel and perpendicular pressures from sound speeds
!
    pr_amb = dens * csnd * csnd / gamma
    pr_ovr = pr_amb * ratio
#endif /* ADI */

! calculate the square of radius
!
    rad    = radius * radius

#ifdef MHD
#ifdef GLM
! set the initial magnetic field
!
    uu(ibx,:,:,:) = bext(inx)
    uu(iby,:,:,:) = bext(iny)
    uu(ibz,:,:,:) = bext(inz)
#else /* GLM */
! set the initial vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    uu(iaz,:,:,:) = 0.0d0
#endif /* GLM */

! preprare magnetic variables
!
    call update_magnetic_variables()
#endif /* MHD */

! set the uniform primitive variables
!
    q(ivx,:) = 0.0d0
    q(ivy,:) = 0.0d0
    q(ivz,:) = 0.0d0

#ifdef MHD
! set the uniform magnetic field
!
    q(ibx,:) = bext(inx)
    q(iby,:) = bext(iny)
    q(ibz,:) = bext(inz)

#ifdef GLM
! set the initial magnetic field
!
    q(iph,:) = 0.0d+00
#endif /* GLM */
#endif /* MHD */

! iterate over all positions in the YZ plane
!
    do k = 1, km

#ifdef R3D
! calculate the corner Z coordinates
!
      zl = abs(z(k)) - dzh
      zu = abs(z(k)) + dzh
#endif /* R3D */

      do j = 1, jm

! calculate the corner Y coordinates
!
        yl = abs(y(j)) - dyh
        yu = abs(y(j)) + dyh

! sweep along the X coordinate
!
        do i = 1, im

! calculate the corner X coordinates
!
          xl = abs(x(i)) - dxh
          xu = abs(x(i)) + dxh

! calculate the minimum and maximum corner distances from the origin
!
#ifdef R3D
          rl = xl * xl + yl * yl + zl * zl
          ru = xu * xu + yu * yu + zu * zu
#else /* R3D */
          rl = xl * xl + yl * yl
          ru = xu * xu + yu * yu
#endif /* R3D */

! set the initial density and pressure
!
          q(idn,i) = dn_amb
#ifdef ADI
          q(ipr,i) = pr_amb
#endif /* ADI */

! set the initial pressure in cells laying completely within the radius
!
          if (ru .le. rad) then

! set the overpressure region density
!
            q(idn,i) = dn_ovr

#ifdef ADI
! set the overpressure region pressure
!
            q(ipr,i) = pr_ovr
#endif /* ADI */

! set the initial pressure in the cell completely outside the radius
!
          else if (rl .ge. rad) then

! set the ambient region density
!
            q(idn,i) = dn_amb

#ifdef ADI
! set the ambient medium pressure
!
            q(ipr,i) = pr_amb
#endif /* ADI */

! integrate density or pressure in cells which are crossed by the circule with
! the given radius
!
          else

#ifdef R3D
! in 3D simply set the ambient values since the integration is more complex
!

! set the ambient region density
!
            q(idn,i) = dn_amb

#ifdef ADI
! set the ambient medium pressure
!
            q(ipr,i) = pr_amb
#endif /* ADI */
#else /* R3D */
! calculate the bounds of area integration
!
            xb = max(xl, sqrt(max(0.0d0, rad - yu * yu)))
            xt = min(xu, sqrt(max(0.0d0, rad - yl * yl)))
            yb = max(yl, sqrt(max(0.0d0, rad - xu * xu)))
            yt = min(yu, sqrt(max(0.0d0, rad - xl * xl)))

! integrate the area below the circle within the current cell for both
! functions, y = f(x) and x = g(y), and then average them to be sure that we
! are getting the ideal symmetry
!
            fc_ovr = 0.5d0 * (rad * (asin(xt / radius) - asin(xb / radius))   &
                   + (xt * yb - xb * yt)) - yl * (xt - xb)
            fc_ovr = fc_ovr + (xb - xl) * dy

            fc_amb = 0.5d0 * (rad * (asin(yt / radius) - asin(yb / radius))   &
                   + (yt * xb - yb * xt)) - xl * (yt - yb)
            fc_amb = fc_amb + (yb - yl) * dx

            fc_ovr = 0.5d0 * (fc_ovr + fc_amb)

! normalize coefficients
!
            fc_ovr = fc_ovr / daxy
            fc_amb = 1.0d0 - fc_ovr

! integrate the density over the edge cells
!
            q(idn,i) = fc_ovr * dn_ovr + fc_amb * dn_amb

#ifdef ADI
! integrate the pressure over the edge cells
!
            q(ipr,i) = fc_ovr * pr_ovr + fc_amb * pr_amb
#endif /* ADI */
#endif /* R3D */

          end if

        end do

! convert the primitive variables to conservative ones
!
        call prim2cons(im, q(1:nt,1:im), u(1:nt,1:im))

! update array of primitive variables
!
        qp(1:nt,1:im,j,k) = q(1:nt,1:im)

! update array of conserved variables
!
        uu(1:nu,1:im,j,k) = u(1:nu,1:im)

      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine blast
!
!===============================================================================
!
! subroutine ROTOR:
! ----------------
!
!   Subroutine sets up the initial conditions for the rotor problem.
!
!   References:
!
!   [1] Balsara, D. S. & Spicer, D. S., "A Staggered Mesh Algorithm Using High
!       Order Godunov Fluxes to Ensure Solenoidal Magnetic Fields in
!       Magnetohydrodynamic Simulations", Journal of Computational Physics,
!       1999, 149, pp. 270-292
!
!===============================================================================
!
  subroutine rotor()

! include external procedures and variables
!
    use constants, only : pi
    use equations, only : prim2cons
#ifdef MHD
    use equations, only : update_magnetic_variables
#endif /* MHD */
    use mesh     , only : im, jm, km
    use mesh     , only : x, y, z, dx, dy, dz, dxh, dyh, dzh, daxy
    use variables, only : nu, nt
    use variables, only : uu, qp
#ifdef MHD
    use variables, only : bext
#endif /* MHD */
    use variables, only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables, only : ipr
#endif /* ADI */
#ifdef MHD
#ifdef GLM
    use variables, only : iph
#endif /* GLM */
    use variables, only : inx, iny, inz
    use variables, only : ibx, iby, ibz
#ifdef VPOT
    use variables, only : iax, iay, iaz
#endif /* VPOT */
#endif /* MHD */

! local variables are not implicit by default
!
    implicit none

! default parameter values
!
    real    :: dens   = 1.00d+0  ! ambient density
#ifdef ADI
    real    :: pres   = 1.00d+0  ! pressure in the whole domain
#endif /* ADI */
    real    :: ratio  = 1.00d+1  ! the ratio of the inner to ambient densities
    real    :: omega  = 2.00d+1  ! the angular speed of the rotor
    real    :: radius = 1.00d-1  ! the rotor radius

! local variables
!
    integer :: i, j, k
    real    :: xl, yl, zl, xu, yu, zu, rl, ru, rd
    real    :: xb, yb, xt, yt
    real    :: fa, fo

! temporary array to store primitive variables
!
    real, dimension(nt,im) :: q, u
!
!------------------------------------------------------------------------------
!
! get problem parameters from the parameter file
!
    call get_parameter_real("dens"  , dens  )
#ifdef ADI
    call get_parameter_real("pres"  , pres  )
#endif /* ADI */
    call get_parameter_real("ratio" , ratio )
    call get_parameter_real("omega" , omega )
    call get_parameter_real("radius", radius)

! prepare squared radius
!
    rd = radius * radius

#if defined MHD && defined VPOT
! set the initial vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    uu(iaz,:,:,:) = 0.0d0

! preprare magnetic variables
!
    call update_magnetic_variables()
#endif /* MHD & VPOT */

! set the uniform primitive variables
!
    q(ivz,:) = 0.0d0
#ifdef ADI
    q(ipr,:) = pres
#endif /* ADI */

#ifdef MHD
! set the uniform magnetic field
!
    q(ibx,:) = bext(inx)
    q(iby,:) = bext(iny)
    q(ibz,:) = bext(inz)

#ifdef GLM
! set the initial div(B) potential
!
    q(iph,:) = 0.0d+00
#endif /* GLM */
#endif /* MHD */

! set the primitive variable arrays
!
    do k = 1, km

#ifdef R3D
! calculate the corner Z coordinates
!
      zl = abs(z(k)) - dzh
      zu = abs(z(k)) + dzh
#endif /* R3D */

      do j = 1, jm

! calculate the corner Y coordinates
!
        yl = abs(y(j)) - dyh
        yu = abs(y(j)) + dyh

        do i = 1, im

! calculate the corner X coordinates
!
          xl = abs(x(i)) - dxh
          xu = abs(x(i)) + dxh

! calculate the minimum and maximum corner distances from the origin
!
#ifdef R3D
          rl = xl * xl + yl * yl + zl * zl
          ru = xu * xu + yu * yu + zu * zu
#else /* R3D */
          rl = xl * xl + yl * yl
          ru = xu * xu + yu * yu
#endif /* R3D */

! set the rotor variables in cells inside the given radius
!
          if (ru <= rd) then

! set the initial density and velocity in the rotor
!
            q(idn,i) =   dens * ratio
            q(ivx,i) = - omega * y(j)
            q(ivy,i) =   omega * x(i)

! set the ambient variables in cells outside the given radius
!
          else if (rl >= rd) then

! set the ambient region density and velocity
!
            q(idn,i) = dens
            q(ivx,i) = 0.0d0
            q(ivy,i) = 0.0d0

! integrate density and velocity in cells which are crossed by the rotor radius
!
          else

#ifdef R3D
! in 3D simply set the ambient values since the integration is more complex
!
            q(idn,i) = dens
            q(ivx,i) = 0.0d0
            q(ivy,i) = 0.0d0
#else /* R3D */
! calculate the bounds of the integration area
!
            xb = max(xl, sqrt(max(0.0d0, rd - yu * yu)))
            xt = min(xu, sqrt(max(0.0d0, rd - yl * yl)))
            yb = max(yl, sqrt(max(0.0d0, rd - xu * xu)))
            yt = min(yu, sqrt(max(0.0d0, rd - xl * xl)))

! integrate the area below the circle within the current cell for both
! functions, y = f(x) and x = g(y), and then average them to be sure that we
! are getting the ideal symmetry
!
            fo = 0.5d0 * (rd * (asin(xt / radius) - asin(xb / radius))         &
                                       + (xt * yb - xb * yt)) - yl * (xt - xb)
            fo = fo + (xb - xl) * dy

            fa = 0.5d0 * (rd * (asin(yt / radius) - asin(yb / radius))         &
                                       + (yt * xb - yb * xt)) - xl * (yt - yb)
            fa = fa + (yb - yl) * dx

            fo = 0.5d0 * (fo + fa)

! normalize coefficients
!
            fo = fo / daxy
            fa = 1.0d0 - fo

! integrate the density and velocity over the edge cells
!
            q(idn,i) = (fo * ratio + fa) * dens
            q(ivx,i) = - omega * y(j) * fo
            q(ivy,i) =   omega * x(i) * fo
#endif /* R3D */

          end if

        end do

! convert the primitive variables to conservative ones
!
        call prim2cons(im, q(1:nt,1:im), u(1:nt,1:im))

! update array of primitive variables
!
        qp(1:nt,1:im,j,k) = q(1:nt,1:im)

! update array of conserved variables
!
#if defined MHD && defined GLM
        uu(1:nt,1:im,j,k) = u(1:nt,1:im)
#else /* MHD & GLM */
        uu(1:nu,1:im,j,k) = u(1:nu,1:im)
#endif /* MHD & GLM */

      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine rotor
!
!===============================================================================
!
! subroutine IMPLOSION:
! --------------------
!
!   Subroutine sets up the initial conditions for implosion test problem.
!
!===============================================================================
!
  subroutine implosion()

! include external procedures and variables
!
    use equations , only : prim2cons
    use mesh      , only : im, jm, km, x, y, z, dxh, dyh, dzh, dx
    use variables , only : nt
    use variables , only : uu, qp
    use variables , only : idn
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */
#ifdef MHD
    use variables , only : inx, iny, inz
    use variables , only : ibx, iby, ibz
#ifdef VPOT
    use variables , only : iax, iay, iaz
    use variables , only : bi
#endif /* VPOT */
    use variables , only : bext
#endif /* MHD */

! local variables are not implicit by default
!
    implicit none

! the parameter determining the shock line: x + y = sline
!
    real    :: sline = 0.5d0

! densities of the ambient and overpressure regions
!
    real    :: odens = 1.0d0, adens = 0.125d0

#ifdef ADI
! pressures of the ambient and overpressure regions
!
    real    :: opres = 1.0d0, apres = 0.140d0
#endif /* ADI */

! local variables
!
    integer :: i, j, k
    real    :: rl, ru, ds, dl, dr

! temporary array to store primitive variables
!
    real, dimension(nt,im) :: q
!
!------------------------------------------------------------------------------
!
! get problem parameters from the parameter file
!
    call get_parameter_real("shock_line" , sline)
    call get_parameter_real("dens_over"  , odens)
    call get_parameter_real("dens"       , adens)
#ifdef ADI
    call get_parameter_real("pres_over"  , opres)
    call get_parameter_real("pres"       , apres)
#endif /* ADI */

! iterate over all positions in the YZ plane
!
    do k = 1, km
      do j = 1, jm

! reset the array of primitive variables
!
        q(:,:) = 0.0d0

! sweep along the X coordinate
!
        do i = 1, im

! calculate cell corner distances from the origin; useful for cell averaging
!
#ifdef R3D
          rl = x(i) + y(j) + z(k) - dxh - dyh - dzh
          ru = x(i) + y(j) + z(k) + dxh + dyh + dzh
#else /* R3D */
          rl = x(i) + y(j) - dxh - dyh
          ru = x(i) + y(j) + dxh + dyh
#endif /* R3D */

! initialize density and pressure in both regions
!
          if (ru .le. sline) then
            q(idn,i) = adens
#ifdef ADI
            q(ipr,i) = apres
#endif /* ADI */
          else if (rl .ge. sline) then
            q(idn,i) = odens
#ifdef ADI
            q(ipr,i) = opres
#endif /* ADI */
          else
            ds = (sline - rl) / dx
            if (ds .le. 1.0d0) then
#ifdef R3D
              dl = 0.5d0 * ds * ds * ds
#else /* R3D */
              dl = 0.5d0 * ds * ds
#endif /* R3D */
              dr = 1.0d0 - dl
            else
              ds = (ru - sline) / dx
#ifdef R3D
              dr = 0.5d0 * ds * ds * ds
#else /* R3D */
              dr = 0.5d0 * ds * ds
#endif /* R3D */
              dl = 1.0d0 - dr
            end if

            q(idn,i) = adens * dl + odens * dr
#ifdef ADI
            q(ipr,i) = apres * dl + opres * dr
#endif /* ADI */
          end if

        end do

#ifdef MHD
! set the initial magnetic field
!
        q(ibx,:) = bext(inx)
        q(iby,:) = bext(iny)
        q(ibz,:) = bext(inz)
#endif /* MHD */

! update array of primitive variables
!
        qp(:,:,j,k) = q(:,:)

! convert the primitive variables to conservative ones
!
        call prim2cons(im, q(:,:), uu(:,:,j,k))

      end do
    end do

#if defined MHD && defined VPOT
! set the face centered magnetic field
!
    bi(inx,:,:,:) = bext(inx)
    bi(iny,:,:,:) = bext(iny)
    bi(inz,:,:,:) = bext(inz)

! reset the vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    uu(iaz,:,:,:) = 0.0d0
#endif /* MHD & VPOT */

!-------------------------------------------------------------------------------
!
  end subroutine implosion
!
!===============================================================================
!
! subroutine KH:
! -------------
!
!   Subroutine sets up the initial conditions for the Kelvin-Helmholtz
!   instability problem.
!
!===============================================================================
!
  subroutine kh()

! include external procedures and variables
!
    use equations , only : prim2cons
    use mesh      , only : im, jm, km, y
    use random    , only : randomn
    use variables , only : nt
    use variables , only : uu, qp
    use variables , only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */
#ifdef MHD
    use variables , only : inx, iny, inz
    use variables , only : ibx, iby, ibz
#ifdef VPOT
    use variables , only : iax, iay, iaz
    use variables , only : bi
#endif /* VPOT */
    use variables , only : bext
#endif /* MHD */

! local variables are not implicit by default
!
    implicit none

! local variables
!
    integer :: i, j, k

! temporary array to store primitive variables
!
    real, dimension(nt,im) :: q
!
!------------------------------------------------------------------------------
!
! iterate over all positions in the YZ plane
!
    do k = 1, km
      do j = 1, jm

! reset the array of primitive variables
!
        q(:,:) = 0.0d0

! set the initial density and the X component of velocity
!
        if (abs(y(j)) .le. 0.25d0) then
          q(idn,:) =   2.0d0
          q(ivx,:) =   0.5d0
        else
          q(idn,:) =   1.0d0
          q(ivx,:) = - 0.5d0
        end if

#ifdef ADI
! set the initial pressure
!
        q(ipr,:) = 2.5d0
#endif /* ADI */

! add a random seed velocity component
!
        do i = 1, im
          q(ivx,i) = q(ivx,i) + 0.01d0 * randomn()
          q(ivy,i) = q(ivy,i) + 0.01d0 * randomn()
#ifndef R2D
          q(ivz,i) = q(ivz,i) + 0.01d0 * randomn()
#endif /* !R2D */
        end do

#ifdef MHD
! set the initial magnetic field
!
        q(ibx,:) = bext(1)
        q(iby,:) = bext(2)
        q(ibz,:) = bext(3)
#endif /* MHD */

! update array of primitive variables
!
        qp(:,:,j,k) = q(:,:)

! convert the primitive variables to conservative ones
!
        call prim2cons(im, q(:,:), uu(:,:,j,k))

      end do
    end do

#if defined MHD && defined VPOT
! set the face centered magnetic field
!
    bi(inx,:,:,:) = bext(inx)
    bi(iny,:,:,:) = bext(iny)
    bi(inz,:,:,:) = bext(inz)

! reset the vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    uu(iaz,:,:,:) = 0.0d0
#endif /* MHD & VPOT */

!-------------------------------------------------------------------------------
!
  end subroutine kh
!
!===============================================================================
!
! subroutine RT:
! -------------
!
!   Subroutine sets up the initial conditions for the Rayleigh-Taylor
!   instability problem.
!
!   Note:
!
!     GRAVITY = Y has to be set in make.config.
!
!===============================================================================
!
  subroutine rt()

! include external procedures and variables
!
    use equations , only : prim2cons
    use mesh      , only : im, jm, km, y
    use random    , only : randomz
    use variables , only : nt
    use variables , only : uu, qp
    use variables , only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */
#ifdef MHD
    use variables , only : inx, iny, inz
    use variables , only : ibx, iby, ibz
#ifdef VPOT
    use variables , only : iax, iay, iaz
    use variables , only : bi
#endif /* VPOT */
    use variables , only : bext
#endif /* MHD */

! local variables are not implicit by default
!
    implicit none

! default parameter values
!
    real    :: dens = 1.0d0, pres = 2.5d0, vper = 0.01d0, gacc = -0.1d0

! local variables
!
    integer :: i, j, k

! temporary array to store primitive variables
!
    real, dimension(nt,im) :: q
!
!------------------------------------------------------------------------------
!
! get problem parameters from the parameter file
!
    call get_parameter_real("dens", dens)
    call get_parameter_real("pres", pres)
    call get_parameter_real("vper", vper)
    call get_parameter_real("gacc", gacc)

! iterate over all positions in YZ plane
!
    do k = 1, km
      do j = 1, jm

! reset the array of primitive variables
!
        q(:,:) = 0.0d0

! set the initial density
!
        if (y(j) .le. 0.0d0) then
          q(idn,:) =         dens
        else
          q(idn,:) = 2.0d0 * dens
        end if

#ifdef ADI
! set the initial pressure
!
        q(ipr,:) = pres + q(idn,:) * gacc * y(j)
#endif /* ADI */

! add a random seed velocity perturbation
!
        do i = 1, im
          q(ivy,i) = vper * randomz()
        end do

#ifdef MHD
! set the initial magnetic field
!
        q(ibx,:) = bext(1)
        q(iby,:) = bext(2)
        q(ibz,:) = bext(3)
#endif /* MHD */

! update array of primitive variables
!
        qp(:,:,j,k) = q(:,:)

! convert the primitive variables to conservative ones
!
        call prim2cons(im, q(:,:), uu(:,:,j,k))

      end do
    end do

#if defined MHD && defined VPOT
! set the face centered magnetic field
!
    bi(inx,:,:,:) = bext(inx)
    bi(iny,:,:,:) = bext(iny)
    bi(inz,:,:,:) = bext(inz)

! reset the vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    uu(iaz,:,:,:) = 0.0d0
#endif /* MHD & VPOT */

!-------------------------------------------------------------------------------
!
  end subroutine rt
!
!===============================================================================
!
! subroutine TURBULENCE:
! ---------------------
!
!   Subroutine sets up the initial conditions for turbulence studies.  The
!   initial conditions are uniform density and pressure, initial velocity is
!   zero everywhere, and the magnetic field is uniform.
!
!===============================================================================
!
  subroutine turbulence()

! include external procedures and variables
!
    use equations, only : prim2cons
    use variables, only : nt
    use variables, only : uu, qp
    use variables, only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables, only : ipr
#endif /* ADI */
#ifdef MHD
#ifdef GLM
    use variables, only : iph
#endif /* GLM */
    use variables, only : inx, iny, inz
    use variables, only : ibx, iby, ibz
#ifdef VPOT
    use variables , only : iax, iay, iaz
    use variables , only : bi
#endif /* VPOT */
    use variables , only : bext
#endif /* MHD */

! local variables are not implicit by default
!
    implicit none

! default parameter values
!
    real    :: dens = 1.0d0
#ifdef ADI
    real    :: pres = 1.0d0
#endif /* ADI */

! local variables
!
    integer :: n

! vector of primitive and conserved variables
!
    real, dimension(nt,1) :: q, u
!
!------------------------------------------------------------------------------
!
! get problem parameters from the parameter file
!
    call get_parameter_real("dens", dens)
#ifdef ADI
    call get_parameter_real("pres", pres)
#endif /* ADI */

! prepare primitive variables
!
    q(idn,:) = dens
    q(ivx,:) = 0.0d0
    q(ivy,:) = 0.0d0
    q(ivz,:) = 0.0d0
#ifdef ADI
    q(ipr,:) = pres
#endif /* ADI */
#ifdef MHD
    q(ibx,:) = bext(inx)
    q(iby,:) = bext(iny)
    q(ibz,:) = bext(inz)
#ifdef GLM
    q(iph,:) = 0.0d+00
#endif /* GLM */
#endif /* MHD */

! convert primitive variables to conservative
!
    call prim2cons(1, q(:,:), u(:,:))

! iterate over all variables
!
    do n = 1, nt

! update the array of primitive variables
!
      qp(n,:,:,:) = q(n,1)

! update the array of conserved variables
!
      uu(n,:,:,:) = u(n,1)

    end do

#if defined MHD && defined VPOT
! set the face centered magnetic field
!
    bi(inx,:,:,:) = bext(inx)
    bi(iny,:,:,:) = bext(iny)
    bi(inz,:,:,:) = bext(inz)

! reset the vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    uu(iaz,:,:,:) = 0.0d0
#endif /* MHD & VPOT */

!-------------------------------------------------------------------------------
!
  end subroutine turbulence
!
#ifdef MHD
!===============================================================================
!
! subroutine OT:
! -------------
!
!   Subroutine sets up the initial conditions for the Orszag-Tang MHD test
!   problem.
!
!   References:
!
!     [1] Orszag & Tang, 1998, J. Fluid Mech. 90, 129 (original version)
!     [2] Miyoshi & Kusano, 2005, J. Comput. Phys., 208, 315 (adiabatic version)
!     [3] Mignone, 2007, J. Comput. Phys., 225, 1427 (isothermal version)
!
!===============================================================================
!
  subroutine ot()

! include external procedures and variables
!
    use constants , only : pi2, pi4
    use equations , only : update_magnetic_variables, prim2cons
#ifdef ISO
    use equations , only : csnd
#endif /* ISO */
#ifdef ADI
    use equations , only : gamma
#endif /* ADI */
    use mesh      , only : im, jm, km, x, y, z, dxh, dyh, dzh, dx, dy, dz
#ifdef GLM
    use operators , only : curl
#endif /* GLM */
    use variables , only : nu, nt
    use variables , only : qp, uu
    use variables , only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */
    use variables , only : ibx, iby, ibz
#ifdef VPOT
    use variables , only : iax, iay, iaz
#endif /* VPOT */
#ifdef GLM
    use variables , only : iph
#endif /* GLM */

! local variables are not implicit by default
!
    implicit none

#ifdef R3D
! runtime parameters
!
    real    :: zeps = 0.2d0
#endif /* R3D */

! local variables
!
    integer :: i, j, k
    real    :: dens, pres, vamp, bamp

! temporary array to store primitive variables
!
    real, dimension(nt,im) :: q, u
    real, dimension(im)    :: xm, xp, xl, xu, sx, cx
    real, dimension(jm)    :: ym, yp, sy, cy
#ifdef R3D
    real, dimension(km)    :: zm, zp, cz, sz
#endif /* R3D */
!
!------------------------------------------------------------------------------
!
! get problem parameters from the parameter file
!
#ifdef R3D
    call get_parameter_real("zeps" , zeps )
#endif /* R3D */

! calculate  density, pressure, and the magnetic field strength
!
#ifdef ISO
    dens = 1.0d0
    vamp = csnd
    bamp = csnd * sqrt(0.6d0) / pi4
#endif /* ISO */
#ifdef ADI
    dens = gamma * gamma
    pres = gamma
    vamp = 1.0d0
    bamp = 1.0d0 / pi4
#endif /* ADI */

! prepare coordinates
!
    xm(:) = pi2 * (x(:) - dxh)
    xp(:) = pi2 * (x(:) + dxh)
    xl(:) = pi4 * (x(:) - dxh)
    xu(:) = pi4 * (x(:) + dxh)
    ym(:) = pi2 * (y(:) - dyh)
    yp(:) = pi2 * (y(:) + dyh)
#ifdef R3D
    zm(:) = pi2 * (z(:) - dzh)
    zp(:) = pi2 * (z(:) + dzh)
#endif /* R3D */

! directional triginometric vectors
!
    do i = 1, im
#if defined CD || defined GLM
      sx(i) = (sin(xu(i)) - sin(xl(i))) / (pi4 * dx)
#endif /* CD | GLM */
#ifdef CT
      sx(i) =  cos(xu(i))
#endif /* CT */
      cx(i) = (cos(xp(i)) - cos(xm(i))) / (pi2 * dx)
    end do
    do j = 1, jm
#if defined CD || defined GLM
      sy(j) = (sin(ym(j)) - sin(yp(j))) / (pi4 * dy)
#endif /* CD | GLM */
#ifdef CT
      sy(j) =  cos(yp(j))
#endif /* CT */
      cy(j) = (cos(ym(j)) - cos(yp(j))) / (pi2 * dy)
    end do
#ifdef R3D
    do k = 1, km
      cz(k) = (cos(zm(k)) - cos(zp(k))) / (pi2 * dz) * zeps
      sz(k) = 1.0d0 + cz(k)
    end do
#endif /* R3D */

#ifdef GLM
! prepare the initial vector potential
!
    uu(ibx,:,:,:) = 0.0d0
    uu(iby,:,:,:) = 0.0d0
    do j = 1, jm
      do i = 1, im
        uu(ibz,i,j,:) = bamp * (sx(i) + 4.0d0 * sy(j))
      end do
    end do

! calculate the initial magnetic field
!
    call curl(im, jm, km, uu(ibx:ibz,:,:,:), qp(ibx:ibz,:,:,:), 'c')
#else /* GLM */
! prepare the initial vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    do j = 1, jm
      do i = 1, im
#ifdef CT
        uu(iaz,i,j,:) = bamp * (sx(i) - 2.0d0 * sy(j))
#endif /* CT */
#ifdef CD
        uu(iaz,i,j,:) = bamp * (sx(i) + 4.0d0 * sy(j))
#endif /* CD */
      end do
    end do

! calculate magnetic field from vector potential
!
    call update_magnetic_variables()
#endif /* GLM */

! iterate over all positions in the YZ plane
!
    do k = 1, km
      do j = 1, jm

! reset the array of primitive variables
!
        q(:,:) = 0.0d+00

! set the initial density
!
        q(idn,:) = dens

#ifdef ADI
! set the initial pressure
!
        q(ipr,:) = pres
#endif /* ADI */

! set the initial velocity
!
#ifdef R3D
        q(ivx,:) = vamp * cy(j) * sz(k)
        q(ivy,:) = vamp * cx(:) * sz(k)
        q(ivz,:) = vamp * cz(k)
#else /* R3D */
        q(ivx,:) = vamp * cy(j)
        q(ivy,:) = vamp * cx(:)
#endif /* R3D */

! set the initial magnetic field
!
        q(ibx,:) = qp(ibx,:,j,k)
        q(iby,:) = qp(iby,:,j,k)
        q(ibz,:) = qp(ibz,:,j,k)

#ifdef GLM
! set the initial magnetic field
!
        q(iph,:) = 0.0d+00
#endif /* GLM */

! convert the primitive variables to conservative ones
!
        call prim2cons(im, q(1:nt,1:im), u(1:nt,1:im))

! update array of primitive variables
!
        qp(1:nt,:,j,k) = q(1:nt,:)

! update array of conserved variables
!
#ifdef GLM
        uu(1:nt,:,j,k) = u(1:nt,:)
#else /* GLM */
        uu(1:nu,:,j,k) = u(1:nu,:)
#endif /* GLM */

      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine ot
!
!===============================================================================
!
! subroutine LOOP:
! ---------------
!
!   Subroutine sets up the initial conditions for the magnetic loop advection
!   test problem.
!
!===============================================================================
!
  subroutine loop()

! include external procedures and variables
!
    use constants , only : d2r
    use equations , only : update_magnetic_variables, prim2cons
    use mesh      , only : im, jm, km
#ifdef CD
    use mesh      , only : x , y
#endif /* CD */
#ifdef CT
    use mesh      , only : xi, yi
#endif /* CT */
#ifdef GLM
    use mesh      , only : x , y
    use operators , only : curl
#endif /* GLM */
    use variables , only : nu, nt
    use variables , only : qp, uu
    use variables , only : idn, ivx, ivy, ivz
#ifdef ADI
    use variables , only : ipr
#endif /* ADI */
#ifdef GLM
    use variables , only : iph
#endif /* GLM */
    use variables , only : ibx, iby, ibz
#ifdef VPOT
    use variables , only : iax, iay, iaz
#endif /* VPOT */

! local variables are not implicit by default
!
    implicit none

! default parameter values
!
    real    :: dens = 1.0d0, pres = 1.0d0, vini = 1.0d0, bini = 1.0d-3
    real    :: radius = 0.3d0, angle = 60.0d0

! local variables
!
    integer :: i, j, k
    real    :: r, ang, vsn, vcs

! temporary array to store primitive variables
!
    real, dimension(nt,im) :: q, u
!
!------------------------------------------------------------------------------
!
! get problem parameters from the parameter file
!
    call get_parameter_real("dens"  , dens  )
    call get_parameter_real("pres"  , pres  )
    call get_parameter_real("vini"  , vini  )
    call get_parameter_real("bini"  , bini  )
    call get_parameter_real("radius", radius)
    call get_parameter_real("angle" , angle )

! velocity component amplitudes
!
    ang = d2r  * angle
    vsn = vini * sin(ang)
    vcs = vini * cos(ang)

#ifdef GLM
! set the initial vector potential
!
    uu(ibx,:,:,:) = 0.0d0
    uu(iby,:,:,:) = 0.0d0
    do j = 1, jm
      do i = 1, im
        r             = sqrt(x (i) * x (i) + y (j) * y (j))
        uu(ibz,i,j,:) = max(0.0d0, bini * (radius - r))
      end do
    end do

! calculate the initial magnetic field
!
    call curl(im, jm, km, uu(ibx:ibz,:,:,:), qp(ibx:ibz,:,:,:), 'c')
#else /* GLM */
! set the initial vector potential
!
    uu(iax,:,:,:) = 0.0d0
    uu(iay,:,:,:) = 0.0d0
    do j = 1, jm
      do i = 1, im
#ifdef CT
        r             = sqrt(xi(i) * xi(i) + yi(j) * yi(j))
#endif /* CT */
#ifdef CD
        r             = sqrt(x (i) * x (i) + y (j) * y (j))
#endif /* CD */
        uu(iaz,i,j,:) = max(0.0d0, bini * (radius - r))
      end do
    end do

! convert vector potential to magnetic field
!
    call update_magnetic_variables()
#endif /* GLM */

! iterate over all positions in the YZ plane
!
    do k = 1, km
      do j = 1, jm

! set the initial density
!
        q(idn,:) = dens

#ifdef ADI
! set the initial pressure
!
        q(ipr,:) = pres
#endif /* ADI */

! set the initial velocity
!
        q(ivx,:) = vcs
        q(ivy,:) = vsn
        q(ivz,:) = 0.0d0

! set the initial magnetic field
!
        q(ibx,:) = qp(ibx,:,j,k)
        q(iby,:) = qp(iby,:,j,k)
        q(ibz,:) = qp(ibz,:,j,k)

#ifdef GLM
! set the initial div(B) potential
!
        q(iph,:) = 0.0d+00
#endif /* GLM */

! convert the primitive variables to conservative ones
!
        call prim2cons(im, q(1:nt,1:im), u(1:nt,1:im))

! update array of primitive variables
!
        qp(1:nt,:,j,k) = q(1:nt,:)

! update array of conserved variables
!
#ifdef GLM
        uu(1:nt,:,j,k) = u(1:nt,:)
#else /* GLM */
        uu(1:nu,:,j,k) = u(1:nu,:)
#endif /* GLM */

      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine loop
#endif /* MHD */

!===============================================================================
!
end module problems
