!!******************************************************************************
!!
!!  This file is part of the GODUNOV source code, a program to perform
!!  Newtonian or relativistic magnetohydrodynamical simulations.
!!
!!  Copyright (C) 2007-2013 Grzegorz Kowal <grzegorz@amuncode.org>
!!
!!  This program is free software: you can redistribute it and/or modify
!!  it under the terms of the GNU General Public License as published by
!!  the Free Software Foundation, either version 3 of the License, or
!!  (at your option) any later version.
!!
!!  This program is distributed in the hope that it will be useful,
!!  but WITHOUT ANY WARRANTY; without even the implied warranty of
!!  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!!  GNU General Public License for more details.
!!
!!  You should have received a copy of the GNU General Public License
!!  along with this program.  If not, see <http://www.gnu.org/licenses/>.
!!
!!******************************************************************************
!!
!! module: SOURCES
!!
!!  This module handles the explicit and implicit source terms.
!!
!!******************************************************************************
!
module sources

#ifdef PROFILE
! include external subroutines
!
  use timers, only : set_timer, start_timer, stop_timer
#endif /* PROFILE */

! module variables are not implicit by default
!
  implicit none

#ifdef PROFILE
! timer indices
!
  integer, save :: isi, isu
#endif /* PROFILE */

! diffusion coefficient for vector potential
!
  real, save :: nu    = 1.0d-03

#ifdef GRAVITY
! grativational acceleration coefficient
!
  real, save :: gacc  = 0.0d+00
#endif /* GRAVITY */

#ifdef VISCOSITY
! the viscosity coefficient
!
  real, save :: visc  = 0.0d+00
#endif /* VISCOSITY */

#ifdef RESISTIVITY
! the magnetic resistivity coefficient
!
  real, save :: ueta  = 0.0d+00

#ifdef ANOMALOUS
! the anomalous magnetic resistivity coefficient
!
  real, save :: aeta  = 0.0d+00

! the critical current density
!
  real, save :: jcrit = 1.0d+16
#endif /* ANOMALOUS */
#endif /* RESISTIVITY */

! by default everything is public
!
  public

!- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
!
  contains
!
!===============================================================================
!
! subroutine INITIALIZE_SOURCES:
! -----------------------------
!
!   Subroutine initializes the source module by reading the module runtime
!   parameters.
!
!===============================================================================
!
  subroutine initialize_sources()

! include external procedures
!
    use mesh      , only : dx_min, dtd
    use parameters, only : get_parameter_real

! local variables are not implicit by default
!
    implicit none

! diffusion coefficient
!
    real :: diff = 1.0d-16
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! set timer descriptions
!
    call set_timer('initialize sources', isi)
    call set_timer('source term update', isu)

! start time accounting for the source terms initialization
!
    call start_timer(isi)
#endif /* PROFILE */

! obtain grativational acceleration coefficient
!
    call get_parameter_real("nu"   , nu   )

#ifdef GRAVITY
! obtain grativational acceleration coefficient
!
    call get_parameter_real("gacc" , gacc )
#endif /* GRAVITY */

#ifdef VISCOSITY
! obtain the viscosity coefficient
!
    call get_parameter_real("visc" , visc )
#endif /* VISCOSITY */

#ifdef RESISTIVITY
! obtain the magnetic resistivity coefficient
!
    call get_parameter_real("ueta" , ueta )

#ifdef ANOMALOUS
! obtain the anomalous magnetic resistivity coefficient
!
    call get_parameter_real("aeta" , aeta )

! obtain the critical current density
!
    call get_parameter_real("jcrit", jcrit)
#endif /* ANOMALOUS */
#endif /* RESISTIVITY */

! calculate the initial diffusion time
!
#ifdef VISCOSITY
    diff = diff + visc
#endif /* VISCOSITY */
#ifdef RESISTIVITY
    diff = diff + ueta
#endif /* RESISTIVITY */
    dtd = dx_min**2 / diff

#ifdef PROFILE
! stop time accounting for the source terms initialization
!
    call stop_timer(isi)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine initialize_sources
#ifdef SOURCES
!
!===============================================================================
!
! subroutine UPDATE_SOURCES:
! -------------------------
!
!   Subroutine adds explicit source terms to the variable increment dU.
!
!===============================================================================
!
  subroutine update_sources()

! include external procedures and variables
!
    use variables , only : uu, du
#ifdef GRAVITY
    use variables , only : idn, imy
#endif /* GRAVITY */
#ifdef ADI
    use variables , only : ien
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none
!
!-------------------------------------------------------------------------------
!
#ifdef PROFILE
! start accounting time for the source terms update
!
    call start_timer(isu)
#endif /* PROFILE */

#ifdef GRAVITY
! add constant gravitational term
!
    if (abs(gacc) .gt. 0.0d0) then
      du(imy,:,:,:) = du(imy,:,:,:) + gacc * uu(idn,:,:,:)
#ifdef ADI
      du(ien,:,:,:) = du(ien,:,:,:) + gacc * uu(imy,:,:,:)
#endif /* ADI */
    end if
#endif /* GRAVITY */

#ifdef VISCOSITY
! add viscosity terms
!
    call diffuse_viscosity()
#endif /* VISCOSITY */

#ifdef MHD
#ifdef RESISTIVITY
! add resistivity diffusive term
!
    call diffuse_resistivity()
#endif /* RESISTIVITY */

#if defined CD && defined R3D
! add diffusive term to the vector potential to prevent unphysical oscillations
! to arise
!
    call diffuse_vpot()
#endif /* CD & R3D */

#ifdef EGLM
! add source terms according to the extended GLM-MHD
!
    call eglm_terms()
#endif /* EGLM */
#endif /* MHD */

#ifdef PROFILE
! stop accounting time for the source terms update
!
    call stop_timer(isu)
#endif /* PROFILE */

!-------------------------------------------------------------------------------
!
  end subroutine update_sources
#ifdef VISCOSITY
!
!===============================================================================
!
! subroutine DIFFUSE_VISCOSITY:
! ----------------------------
!
!   Subroutine adds a diffusive term corresponding to viscosity.
!
!===============================================================================
!
  subroutine diffuse_viscosity()

! include external procedures and variables
!
    use mesh     , only : im, jm, km
    use mesh     , only : dxi, dyi, dzi
    use operators, only : divergence
    use variables, only : uu, du, qp
    use variables, only : idn, imx, imy, imz, ivx, ivy, ivz
#ifdef ADI
    use variables, only : ien
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! position indices
!
    integer       :: i, j, k
    integer       :: im1, jm1, km1, ip1, jp1, kp1

! saved variables
!
    logical, save :: first = .true.
    real   , save :: nudxi , nudyi , nudzi
    real   , save :: nudx2i, nudy2i, nudz2i

! local variables
!
    real          :: dvx, dvy, dvz, dvg, dvv

! temporary arrays
!
    real, dimension(im,jm,km) :: dv
!
!-------------------------------------------------------------------------------
!
! if the viscosity coefficient is not positive, exit from the subroutine
! without adding anything
!
    if (visc <= 0.0d+00) return

    if (first) then

! set coefficients for viscose terms
!
      nudxi  = visc * dxi / 6.0d+00
      nudyi  = visc * dyi / 6.0d+00
      nudzi  = visc * dzi / 6.0d+00
      nudz2i = visc * dzi**2
      nudx2i = visc * dxi**2
      nudy2i = visc * dyi**2
      nudz2i = visc * dzi**2

! reset the first execution flag
!
      first = .false.

    end if

! calculate the divergence of velocity
!
    call divergence(im, jm, km, qp(ivx:ivz,:,:,:), dv(:,:,:), "c")

! calculate the viscosity term
!
    do k = 1, km
#ifdef R3D
      km1 = max(k - 1, 1)
      kp1 = min(k + 1,km)
#endif /* R3D */
      do j = 1, jm
        jm1 = max(j - 1, 1)
        jp1 = min(j + 1,jm)
        do i = 1, im
          im1 = max(i - 1, 1)
          ip1 = min(i + 1,im)

! directional second derivatives of Vx
!
          dvx = (qp(ivx,ip1,j,k) + qp(ivx,im1,j,k)) - 2.0d+00 * qp(ivx,i,j,k)
          dvy = (qp(ivx,i,jp1,k) + qp(ivx,i,jm1,k)) - 2.0d+00 * qp(ivx,i,j,k)
#ifdef R3D
          dvz = (qp(ivx,i,j,kp1) + qp(ivx,i,j,km1)) - 2.0d+00 * qp(ivx,i,j,k)
#endif /* R3D */

! contribution from the X derivative of divergence
!
          dvg = nudxi * (dv(ip1,j,k) - dv(im1,j,k))

! update the X momentum increment
!
#ifdef R3D
          dvv = nudx2i * dvx + nudy2i * dvy + nudz2i * dvz + dvg
#else /* R3D */
          dvv = nudx2i * dvx + nudy2i * dvy + dvg
#endif /* R3D */
          du(imx,i,j,k) = du(imx,i,j,k) + dvv

#ifdef ADI
! add energy dissipation due to the vescosity term
!
          du(ien,i,j,k) = du(ien,i,j,k) + qp(ivx,i,j,k) * dvv
#endif /* ADI */

! directional second derivatives of Vy
!
          dvx = (qp(ivy,ip1,j,k) + qp(ivy,im1,j,k)) - 2.0d+00 * qp(ivy,i,j,k)
          dvy = (qp(ivy,i,jp1,k) + qp(ivy,i,jm1,k)) - 2.0d+00 * qp(ivy,i,j,k)
#ifdef R3D
          dvz = (qp(ivy,i,j,kp1) + qp(ivy,i,j,km1)) - 2.0d+00 * qp(ivy,i,j,k)
#endif /* R3D */

! contribution from the Y derivative of divergence
!
          dvg = nudyi * (dv(i,jp1,k) - dv(i,jm1,k))

! update the Y momentum increment
!
#ifdef R3D
          dvv = nudx2i * dvx + nudy2i * dvy + nudz2i * dvz + dvg
#else /* R3D */
          dvv = nudx2i * dvx + nudy2i * dvy + dvg
#endif /* R3D */
          du(imy,i,j,k) = du(imy,i,j,k) + dvv

#ifdef ADI
! add energy dissipation due to the vescosity term
!
          du(ien,i,j,k) = du(ien,i,j,k) + qp(ivx,i,j,k) * dvv
#endif /* ADI */

#ifndef R2D
! directional second derivatives of Vz
!
          dvx = (qp(ivz,ip1,j,k) + qp(ivz,im1,j,k)) - 2.0d+00 * qp(ivz,i,j,k)
          dvy = (qp(ivz,i,jp1,k) + qp(ivz,i,jm1,k)) - 2.0d+00 * qp(ivz,i,j,k)
#ifdef R3D
          dvz = (qp(ivz,i,j,kp1) + qp(ivz,i,j,km1)) - 2.0d+00 * qp(ivz,i,j,k)

! contribution from the Y derivative of divergence
!
          dvg = nudzi * (dv(i,j,kp1) - dv(i,j,km1))
#endif /* R3D */

! update the Z momentum increment
!
#ifdef R3D
          dvv = nudx2i * dvx + nudy2i * dvy + nudz2i * dvz + dvg
#else /* R3D */
          dvv = nudx2i * dvx + nudy2i * dvy
#endif /* R3D */
          du(imz,i,j,k) = du(imz,i,j,k) + dvv

#ifdef ADI
! add energy dissipation due to the vescosity term
!
          du(ien,i,j,k) = du(ien,i,j,k) + qp(ivx,i,j,k) * dvv
#endif /* ADI */
#endif /* !R2D */

        end do
      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine diffuse_viscosity
#endif /* VISCOSITY */
#ifdef MHD
#ifdef RESISTIVITY
!
!===============================================================================
!
! subroutine DIFFUSE_RESISTIVITY:
! ------------------------------
!
!   Subroutine adds a diffusive term corresponding to resistivity.
!
!===============================================================================
!
  subroutine diffuse_resistivity()

! include external procedures and variables
!
    use mesh     , only : im, jm, km
#ifdef VPOT
    use operators, only : atoj
#endif /* VPOT */
    use operators, only : curl
#ifdef ADI
    use operators, only : divergence
#endif /* ADI */
    use variables, only : du, qp
#ifdef VPOT
    use variables, only : uu
#endif /* VPOT */
    use variables, only : inx, iny, inz
    use variables, only : ibx, iby, ibz
#ifdef VPOT
    use variables, only : iax, iay, iaz
#endif /* VPOT */
    use variables, only : ijx, ijy, ijz
#ifdef ANOMALOUS
    use variables, only : iet
#endif /* ANOMALOUS */
#ifdef ADI
    use variables, only : ien
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! position indices
!
    integer       :: i, j, k
#if defined ADI && defined CT
    integer       :: im1, jm1, km1
    integer       :: ip1, jp1, kp1
#endif /* ADI & CT */

! saved variables
!
    logical, save :: first = .true.
    real   , save :: eta

! local variables
!
#ifdef ADI
#ifdef CT
    real          :: bx, by, bz, ejx, ejy, ejz
#endif /* CT */
#endif /* ADI */
#ifdef ANOMALOUS
    real          :: ja, xc, fc
#endif /* ANOMALOUS */

! temporary arrays
!
    real, dimension(3,im,jm,km) :: ej, cej
!
!-------------------------------------------------------------------------------
!
! if the resistivity coefficients are not positive, exit from the subroutine
! without doing anything
!
#ifdef ANOMALOUS
    if (ueta <= 0.0d+00 .and. aeta <= 0.0d+00) return
#else /* ANOMALOUS */
    if (ueta <= 0.0d+00) return
#endif /* ANOMALOUS */

! prepare the coefficients which do not change
!
    if (first) then

! set eta to uniform resistivity
!
      eta   = ueta

! reset the first execution flag
!
      first = .false.

    end if

! calculate current density J = ∇xB
!
#ifdef VPOT
    call atoj(im, jm, km, uu(iax:iaz,:,:,:), qp(ijx:ijz,:,:,:))
#endif /* VPOT */
#ifdef GLM
    call curl(im, jm, km, qp(ibx:ibz,:,:,:), qp(ijx:ijz,:,:,:), 'c')
#endif /* GLM */

! prepare local resistivity and calculate (η J)
!
    do k = 1, km
      do j = 1, jm
        do i = 1, im

#ifdef ANOMALOUS
! calculate the absolute value of current density
!
#ifdef R2D
          ja  = abs(qp(ijz,i,j,k))
#else /* R2D */
          ja  = sqrt(sum(qp(ijx:ijz,i,j,k) * qp(ijx:ijz,i,j,k)))
#endif /* R2D */

! calculate the local resistivity (uniform + anomalous)
!
          xc  = ja / jcrit
          fc  = 0.5d+00 * (tanh(20.0d+00 * (xc - 1.0d+00)) + 1.0d+00)
          eta = ueta + aeta * fc * xc

! store the resistivity in an array
!
          qp(iet,i,j,k) = eta
#endif /* ANOMALOUS */

! multiply the current density by the local resistivity
!
          ej(inx,i,j,k) = eta * qp(ijx,i,j,k)
          ej(iny,i,j,k) = eta * qp(ijy,i,j,k)
          ej(inz,i,j,k) = eta * qp(ijz,i,j,k)

        end do
      end do
    end do

#ifdef GLM
! calculate curl of (η J)
!
    call curl(im, jm, km, ej(inx:inz,:,:,:), cej(inx:inz,:,:,:), 'c')

! update magnetic field components
!
    du(ibx,:,:,:) = du(ibx,:,:,:) - cej(inx,:,:,:)
    du(iby,:,:,:) = du(iby,:,:,:) - cej(iny,:,:,:)
    du(ibz,:,:,:) = du(ibz,:,:,:) - cej(inz,:,:,:)
#else /* GLM */
! update vector potential components
!
    du(iax,:,:,:) = du(iax,:,:,:) -  ej(inx,:,:,:)
    du(iay,:,:,:) = du(iay,:,:,:) -  ej(iny,:,:,:)
    du(iaz,:,:,:) = du(iaz,:,:,:) -  ej(inz,:,:,:)
#endif /* GLM */

#ifdef ADI
#ifdef CT
! iterate over all cells
!
    do k = 1, km
      km1 = max( 1, k - 1)
      kp1 = min(km, k + 1)
      do j = 1, jm
        jm1 = max( 1, j - 1)
        jp1 = min(jm, j + 1)
        do i = 1, im
          im1 = max( 1, i - 1)
          ip1 = min(im, i + 1)

! prepare variable [B x (η J)] used in the resistive term for energy equation
!
#ifdef R2D
! interpolate magnetic field and current density at location (i+1/2,j,k)
!
          by  = qp(iby,i,j,k) + qp(iby,ip1,j,k)
          ejz = ej(inz,i,j,k) + ej(inz,i,jm1,k)

! calculate the X component of [B x (η J)]
!
          cej(inx,i,j,k) =   0.25d+00 * by * ejz

! interpolate magnetic field and current density at location (i,j+1/2,k)
!
          bx  = qp(ibx,i,j,k) + qp(ibx,i,jp1,k)
          ejz = ej(inz,i,j,k) + ej(inz,im1,j,k)

! calculate the Y component of [B x (η J)]
!
          cej(iny,i,j,k) = - 0.25d+00 * bx * ejz
#else /* R2D */
! interpolate magnetic field and current density at location (i+1/2,j,k)
!
          by  = qp(iby,i,j,k) + qp(iby,ip1,j,k)
          bz  = qp(ibz,i,j,k) + qp(ibz,ip1,j,k)
          ejy = ej(iny,i,j,k) + ej(iny,i,j,km1)
          ejz = ej(inz,i,j,k) + ej(inz,i,jm1,k)

! calculate the X component of [B x (η J)]
!
          cej(inx,i,j,k) = 0.25d+00 * (by * ejz - bz * ejy)

! interpolate magnetic field and current density at location (i,j+1/2,k)
!
          bx  = qp(ibx,i,j,k) + qp(ibx,i,jp1,k)
          bz  = qp(ibz,i,j,k) + qp(ibz,i,jp1,k)
          ejx = ej(inx,i,j,k) + ej(inx,i,j,km1)
          ejz = ej(inz,i,j,k) + ej(inz,im1,j,k)

! calculate the Y component of [B x (η J)]
!
          cej(iny,i,j,k) = 0.25d+00 * (bz * ejx - bx * ejz)

! interpolate magnetic field and current density at location (i,j,k+1/2)
!
          bx  = qp(ibx,i,j,k) + qp(ibx,i,j,kp1)
          by  = qp(iby,i,j,k) + qp(iby,i,j,kp1)
          ejx = ej(inx,i,j,k) + ej(inx,i,jm1,k)
          ejy = ej(iny,i,j,k) + ej(iny,im1,j,k)

! calculate the Z component of [B x (η J)]
!
          cej(inz,i,j,k) = 0.25d+00 * (bx * ejy - by * ejx)
#endif /* R2D */
        end do
      end do
    end do

! calculate the divergence of B x (η J), i.e. ∇.[B x (η J)]
!
    call divergence(im, jm, km, cej(inx:inz,:,:,:), ej(inx,:,:,:), 'f')
#endif /* CT */

#if defined CD || defined GLM
! calculate B x (η J)
!
    cej(inx,:,:,:) = qp(iby,:,:,:) * ej(inz,:,:,:)                              &
                                               - qp(ibz,:,:,:) * ej(iny,:,:,:)
    cej(iny,:,:,:) = qp(ibz,:,:,:) * ej(inx,:,:,:)                              &
                                               - qp(ibx,:,:,:) * ej(inz,:,:,:)
    cej(inz,:,:,:) = qp(ibx,:,:,:) * ej(iny,:,:,:)                              &
                                               - qp(iby,:,:,:) * ej(inx,:,:,:)

! calculate the divergence of B x (η J), i.e. ∇.[B x (η J)]
!
    call divergence(im, jm, km, cej(inx:inz,:,:,:), ej(inx,:,:,:), 'c')
#endif /* CD | GLM */

! add the resistive source term to the energy equation, i.e.
!     d/dt E + ∇.F = ∇.(B x (η J))
!
    du(ien,:,:,:) = du(ien,:,:,:) + ej(inx,:,:,:)
#endif /* ADI */

!-------------------------------------------------------------------------------
!
  end subroutine diffuse_resistivity
#endif /* RESISTIVITY */
#if defined CD && defined R3D
!
!===============================================================================
!
! subroutine DIFFUSE_VPOT:
! -----------------------
!
!   Subroutine adds a small diffusive term to vector potential in order to
!   prevent some unphysical oscilations to arise.
!
!   References:
!
!     [1] Helzel, Rossmanith, & Taetz, 2011, J. Comput. Phys., 230, 3803
!
!===============================================================================
!
  subroutine diffuse_vpot()

! include external procedures and variables
!
    use mesh     , only : im, jm, km
    use mesh     , only : dt
    use variables, only : uu, du
    use variables, only : iax, iay, iaz

! local variables are not implicit by default
!
    implicit none

! local indices
!
    integer :: i, j, k
    integer :: im1, jm1, km1, ip1, jp1, kp1

! local coefficients
!
    real    :: nudt, dap, dam, da2, ap, am, as, aa, bp, bm, bs

! local parameters
!
    real, parameter :: eps = epsilon(bs)
!
!-------------------------------------------------------------------------------
!
! if the diffusion coefficient is non positive or zero, skip the diffusion term
!
    if (nu .le. 0.0d0) return

! prepare coefficient (ν / Δt)
!
    nudt = nu / dt

! process all positions
!
    do k = 1, im
      km1 = max( 1, k - 1)
      kp1 = min(km, k + 1)
      do j = 1, jm
        jm1 = max( 1, j - 1)
        jp1 = min(jm, j + 1)
        do i = 1, im
          im1 = max( 1, i - 1)
          ip1 = min(im, i + 1)

! === process Ax ===
!
! calculate the left and right derivatives
!
          dap = uu(iax,ip1,j,k) - uu(iax,i,j,k)
          dam = uu(iax,im1,j,k) - uu(iax,i,j,k)

! calculate the second derivative
!
          da2 = dap + dam

! calculate the left and right smoothness coefficients
!
          bm  = dam * dam
          bp  = dap * dap
          bs  = abs(bp - bm)
          am  = 1.0d0 + bs / (bm + eps)
          ap  = 1.0d0 + bs / (bp + eps)
          as  = ap + am

! calculate the smoothness indicator
!
          aa  = max(abs(am / as), abs(ap / as))

! add diffusive term to the Ax increment
!
          du(iax,i,j,k) = du(iax,i,j,k) + (nudt * aa * da2)

! === process Ay ===
!
! calculate the left and right derivatives
!
          dap = uu(iay,i,jp1,k) - uu(iay,i,j,k)
          dam = uu(iay,i,jm1,k) - uu(iay,i,j,k)

! calculate the second derivative
!
          da2 = dap + dam

! calculate the left and right smoothness coefficients
!
          bm  = dam * dam
          bp  = dap * dap
          bs  = abs(bp - bm)
          am  = 1.0d0 + bs / (bm + eps)
          ap  = 1.0d0 + bs / (bp + eps)
          as  = ap + am

! calculate the smoothness indicator
!
          aa  = max(abs(am / as), abs(ap / as))

! add diffusive term to the Ax increment
!
          du(iay,i,j,k) = du(iay,i,j,k) + (nudt * aa * da2)

#ifdef R3D
! === process Az ===
!
! calculate the left and right derivatives
!
          dap = uu(iaz,i,j,kp1) - uu(iaz,i,j,k)
          dam = uu(iaz,i,j,km1) - uu(iaz,i,j,k)

! calculate the second derivative
!
          da2 = dap + dam

! calculate the left and right smoothness coefficients
!
          bm  = dam * dam
          bp  = dap * dap
          bs  = abs(bp - bm)
          am  = 1.0d0 + bs / (bm + eps)
          ap  = 1.0d0 + bs / (bp + eps)
          as  = ap + am

! calculate the smoothness indicator
!
          aa  = max(abs(am / as), abs(ap / as))

! add diffusive term to the Ax increment
!
          du(iaz,i,j,k) = du(iaz,i,j,k) + (nudt * aa * da2)
#endif /* R3D */
        end do
      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine diffuse_vpot
#endif /* CD & R3D */
#ifdef EGLM
!
!===============================================================================
!
! subroutine EGLM_TERMS:
! ---------------------
!
!   Subroutine adds source terms to the momentum and energy equation according
!   to the extended GLM-MHD equations.
!
!   References:
!
!     [1] Dedner, A., Kemm, F., Kröner, D., Munz, C.-D., Schnitzer, T., &
!         Wesenberg, M., "Hyperbolic Divergence Cleaning for the MHD Equations",
!         Journal of Computational Physics, 2002, 175, 645-673
!
!===============================================================================
!
  subroutine eglm_terms()

! include external procedures and variables
!
    use mesh     , only : im, jm, km
    use mesh     , only : dxih, dyih, dzih
    use variables, only : qp, du
    use variables, only : imx, imy, imz, ibx, iby, ibz, iph
#ifdef HEGLM
    use variables, only : ivx, ivy, ivz
#endif /* HEGLM */
#ifdef ADI
    use variables, only : ien
#endif /* ADI */

! local variables are not implicit by default
!
    implicit none

! local indices
!
    integer :: i, j, k, im1, jm1, km1, ip1, jp1, kp1
    real    :: dbx, dby, dbz, divb, dphx, dphy, dphz, vb
!
!-------------------------------------------------------------------------------
!
! add EGLM-MHD source terms due to the divergence of B and potential ψ
!
    do k = 1, km
#ifdef R3D
      km1 = max( 1, k - 1)
      kp1 = min(km, k + 1)
#endif /* R3D */
      do j = 1, jm
        jm1 = max( 1, j - 1)
        jp1 = min(jm, j + 1)
        do i = 1, im
          im1 = max( 1, i - 1)
          ip1 = min(im, i + 1)

! calculate the divergence of magnetic field
!
          dbx  = dxih * (qp(ibx,ip1,j,k) - qp(ibx,im1,j,k))
          dby  = dyih * (qp(iby,i,jp1,k) - qp(iby,i,jm1,k))
#ifdef R3D
          dbz  = dzih * (qp(ibz,i,j,kp1) - qp(ibz,i,j,km1))
          divb = dbx + dby + dbz
#else /* R3D */
          divb = dbx + dby
#endif /* R3D */

! update momentum equations with source term (∇.B)B
!
          du(imx,i,j,k) = du(imx,i,j,k) - divb * qp(ibx,i,j,k)
          du(imy,i,j,k) = du(imy,i,j,k) - divb * qp(iby,i,j,k)
          du(imz,i,j,k) = du(imz,i,j,k) - divb * qp(ibz,i,j,k)

#ifdef HEGLM
! update magnetic field components with source term (∇.B)v
!
          du(ibx,i,j,k) = du(ibx,i,j,k) - divb * qp(ivx,i,j,k)
          du(iby,i,j,k) = du(iby,i,j,k) - divb * qp(ivy,i,j,k)
          du(ibz,i,j,k) = du(ibz,i,j,k) - divb * qp(ivz,i,j,k)
#endif /* HEGLM */

#if defined HEGLM && defined ADI
! calculate scalar product of velocity and magnetic field
!
          vb = sum(qp(ivx:ivz,i,j,k) * qp(ibx:ibz,i,j,k))

! update energy equation
!
          du(ien,i,j,k) = du(ien,i,j,k) - divb * vb
#endif /* HEGLM & ADI */

#if defined HEGLM || defined ADI
! calculate the gradient of potential ψ
!
          dphx = dxih * (qp(iph,ip1,j,k) - qp(iph,im1,j,k))
          dphy = dyih * (qp(iph,i,jp1,k) - qp(iph,i,jm1,k))
#ifdef R3D
          dphz = dzih * (qp(iph,i,j,kp1) - qp(iph,i,j,km1))
#endif /* R3D */
#endif /* HEGLM | ADI */

#ifdef ADI
! update the energy equation with source term B.(∇ψ)
!
          du(ien,i,j,k) = du(ien,i,j,k) - qp(ibx,i,j,k) * dphx                 &
#ifdef R3D
                        - qp(iby,i,j,k) * dphy - qp(ibz,i,j,k) * dphz
#else /* R3D */
                        - qp(iby,i,j,k) * dphy
#endif /* R3D */
#endif /* ADI */

#ifdef HEGLM
! update divergence equation
!
          du(iph,i,j,k) = du(iph,i,j,k) - qp(ivx,i,j,k) * dphx                 &
#ifdef R3D
                        - qp(ivy,i,j,k) * dphy - qp(ivz,i,j,k) * dphz
#else /* R3D */
                        - qp(ivy,i,j,k) * dphy
#endif /* R3D */
#endif /* HEGLM */

        end do
      end do
    end do

!-------------------------------------------------------------------------------
!
  end subroutine eglm_terms
#endif /* EGLM */
#endif /* MHD */
#endif /* SOURCES */

!===============================================================================
!
end module sources
